# Global Sea Level Composition (GSLcomp)

## Installation

We recommend to create a new anaconda environment 

    conda create -n gslcomp python=3 matplotlib numpy scipy pathlib pandas notebook geopandas pytest netcdf4 func_timeout geoplot pytables xarray cf-units dask openpyxl bottleneck numba

package version requirements:
numpy>=1.21.1
pandas>=1.3.1

Activate the environment

    conda activate gslcomp

To install the gslcomp package, go to the root of the package run

    pip install -e .

## Input Data

Follow the steps below to set up the necessary input data:

### 1. Setting up Directories
In the root directory of `gslcomp`, create an input directory:
```bash
mkdir input
```

### 2. Downloading and Unzipping `CoDEC.zip`
Download `CoDEC.zip` from [Zenodo](https://zenodo.org/deposit/8322750) into the `input` directory.
After downloading, unzip the file within the `input` directory:
```bash
cd input
unzip CoDEC.zip
```

### 3. Creating Symlinks
Navigate to the `input` directory and create symlinks:
```bash
ln -s CoDEC/cf_esl gtsm_tide_surge_slr
ln -s CoDEC/cf_tides gtsm_tide_slr
```

### 4. Downloading Additional Files
- **HR File**: Download `HybridRec_Upd0422.mat` from [Zenodo](https://zenodo.org/deposit/8322750) and place it in the `input` directory.
  
- **GESLA3**: Download data from [GESLA3](https://gesla.org) and place it in a directory called `gesla3` inside `input`.

- **PSMSL Data**: Download the `rlr_annual` and `rlr_monthly` datasets from [PSMSL](https://psmsl.org/data/obtaining/year_end/2022/) and place it in the `input` directory.

- **MDT File**: Download the mean dynamic topography (MDT) file from [this link](https://doi.org/10.48670/moi-00150) and place it in the `input` directory.

## Preprocessing

To ensure the compatibility and accuracy of the diverse data sources, a sequence of preprocessing steps is vital before the model execution. 

### Running Preprocessing Scripts

All preprocessing scripts can be initiated via the command:

```bash
python PREPROCESSING_SCRIPT.py --name yourname
```

Additionally, bash scripts are provided in the `slurm_scripts` directory to assist with the preprocessing. Though these scripts are optimized for the slurm job scheduler, they're adaptable for direct bash execution. Remember to modify the paths as necessary.

### Sequence of Preprocessing

Execute the preprocessing steps in the following order:

1. **GTSM Preprocessing**
   
   ```bash
   sh slurm_scripts/run_gtsm_preprocessing.sh
   ```

   - This step generates an index file for all GTSM locations. This index lists the filenames and tags them with a "coastal" flag.

2. **HR Preprocessing**
   
   ```bash
   sh slurm_scripts/run_hr_preprocessing.sh
   ```

   - This script processes the `HybridRec_Upd0422.mat` file, extracting monthly values and saving them as individual h5 files for every GTSM station. 
   
   - Built-in assertions validate that the latitude and longitude of an entry in the .mat file match the associated GTSM filename. If an assertion fails, reconfirm that the GTSM and HR runs correspond to identical locations. A mismatch between the GTSM filenames and entries in the .mat file might necessitate a more sophisticated indexing approach, rooted in their coordinates.

## Run the model
Run the model with

  python main.py --name RUN_ID

or with `slurm_scripts/run_main.sh`

### Optional parameters

* --input_dir
  * location of input data. Default is BASE_DIR/input
* --output_base
  * output will be saved in a subdirectory to the directory specified here. Default is BASE_DIR/output
* --overwrite 
  * if specified, previously saved data using the same `--name` will be overwritten
* --timeout
  * maximum number of seconds to compute one location

## Postprocessing

To reproduce the figures and analyses as showcased in Treu et al., 2023, it's essential to follow a sequence of postprocessing steps. All the necessary bash scripts for postprocessing are in the `slurm_scripts` directory and must be executed in a specific order.

**Note:** Ensure that your system has at least 120 GB RAM, as some postprocessing scripts are resource-intensive.

### Sequence of Postprocessing

Execute the scripts in the following order:

1. **Create Validation Dataset for PSMSL**:
   
   ```bash
   sh slurm_scripts/run_create_val_ds_psmsl.sh
   ```

2. **Create Selected Validation Dataset for PSMSL**:
   
   ```bash
   sh slurm_scripts/run_create_val_ds_psmsl_selected.sh
   ```

3. **Create Validation Dataset for GESLA**:
   
   ```bash
   sh slurm_scripts/run_create_val_ds_gesla.sh
   ```

4. **Create Validation Dataset for Extremes**:
   
   ```bash
   sh slurm_scripts/run_create_val_ds_extremes.sh
   ```

5. **TG Prediction**:
   
   ```bash
   sh slurm_scripts/run_tg_pred.sh
   ```

6. **Merge Predicted Tides**:
   
   ```bash
   sh slurm_scripts/run_merge_predicted_tides.sh
   ```

7. **Create Extremes**:
   
   ```bash
   sh slurm_scripts/run_create_extremes.sh
   ```

8. **Create hcc_gesla3_water_level_hourly_2011_2015**:
   ```bash
   sh slurm_scripts/run_create_hcc_gesla3_water_level_hourly_2011_2015.sh
   ```

After completing the above sequence, your results should align with the findings presented in Treu et al., 2023.

  
## Parallelization with slurm

If slurm is used, jobs are distributed on all available array-tasks

Use e.g.

  `#SBATCH --array=0-1872%800`

with this setting, each job on the cluster combines data at 10 locations.

## Notebooks 
The `notebooks` subdirectory contains jupyter-notebooks to produce figures from Treu et al. 2023. 

* Fig. 1 : `psmsl_based_analysis.ipynb`
* Fig. 2: `tide_gauge_map.ipynb`
* Fig. 3 & 4 : `gesla_based_analysis_fig3_fig4.ipynb`
* Fig.5 : `extremes_analysis_figure_5.ipynb`
* Table1 : `psmsl_based_analysis.ipynb`, `gesla_based_analysis_fig3_fig4.ipynb`, `extremes_analysis_figure_5.ipynb`

* Fig. S1: `sensitivity_checks-autocorrelation`
* Fig. S2: `trend_estimation_sensitivity`


## Dry Locations
There are two locations in the GTSM dataset where the waterlevel is <= the bedlevel:
  * `gtsm_station07568` at lat = 75.015, lon = -92.197
    * bedlevel = 14.25 m, waterlevel == bedlevel for all locations
  * `gtsm_station08226` at lat = 65.024, lon = 179.971
    * bedlevel = -999 m, waterlevel < bedlevel for all locations

These locations are reported in the `error.log` when you run `main.py` 
