"""
containing the main logic of the project
"""
from pathlib import Path

import cftime
import numpy as np
import pandas as pd
import psutil
import statsmodels.api as sm
import xarray as xr
from cf_units import Unit
from pytest import approx

from gslcomp.data_loading import load_gtsm_output


def lin_regr(timeseries):
    """

    Parameters
    ----------
    timeseries: xr.DataArray
        monthly timeseries with 12 months for every year

    Returns
    -------
    intercept, slope, x: float, float, np.Array
        intercept [mm] and slope [mm/year] for the linear fit on the timeseries and the time index x
    """
    # The output should be in mm/year.
    # Therefor x increases by 1 for every year and evenly splits the rest of the year.
    if len(timeseries) != len(timeseries.resample(time="1a").mean()) * 12:
        raise ValueError("timeseries should be a monthly dataset with full years")
    x = (
        timeseries.to_dataframe().index.year
        + (timeseries.to_dataframe().index.month - 1) / 12
    ).to_numpy()
    X = np.expand_dims(x, axis=1)
    X = sm.add_constant(X)
    model = sm.OLS(timeseries.to_dataframe()[timeseries.name], X, missing="drop")
    results = model.fit()
    return results.params.const, results.params.x1, x


class RegressionDataset:
    """
    Provides a time index of type float to be used in a linear regression.
    Also provides a normalized time series and a rescaling method (inverse of the normalization).
    """

    def __init__(self, timeseries):
        """
        Initialize the regressionDataset and compute normalized time series and t, t^2 indecies with
        the udunits unit year.
        The normalization includes shifting the time series such that it has a mean of 0
        over the first 6 years and scaling it to have a standard deviation of 1.

        Parameters
        ----------
        timeseries: xr.DataArray
        """
        self.offset = timeseries.to_dataframe()[timeseries.name].first("6a").mean()
        self.std = timeseries.std().item()
        self.normalized_ts = xr.DataArray(
            dims="time",
            data=self.normalize(timeseries),
            attrs={"long_name": "Normalized time series", "units": timeseries.units,},
            coords=timeseries.coords,
        )
        first_day = timeseries.time[0].dt
        hours_since_start = cftime.date2num(
            timeseries.to_dataframe().index.to_pydatetime(),
            f"hours since {first_day.year.item()}-{first_day.month.item()}-{first_day.day.item()} 00:00:00",
        )

        # t is the time index as float with years since 1900-01-01 00:00
        self.t = xr.DataArray(
            dims="time",
            data=Unit("hour").convert(hours_since_start, Unit("year")),
            coords=timeseries.coords,
        )
        self.t_square = self.t ** 2

    def normalize(self, timeseries):
        """
        Shift and scale time series
        Parameters
        ----------
        timeseries: xr.DataArray

        Returns
        -------

        """
        return (timeseries - self.offset) / self.std

    def rescale(self, timeseries):
        """
        Inverse of normalization
        Parameters
        ----------
        timeseries: xr.DataArray

        Returns
        -------
        xr.DataArray
        """
        return timeseries * self.std + self.offset


def highpass_filter(surge, window_size):
    """
    Highpass filter surge time series using a moving average filter.
    Parameters
    ----------
    surge : xr.DataArray
        time series of waterlevel containing only surges
    window_size : int
        window size for moving average filter
    Returns
    -------
    xr.DataArray
        time series of waterlevel containing surges without variability > 1 month
    """
    # set min_periods=window_size//2 to get lowpass filtered data for the complete period.
    # This might bias the beginning end of the time series.
    low_pass = surge.rolling(
        time=window_size, min_periods=window_size // 2, center=True
    ).mean()
    return xr.DataArray(
        dims="time",
        data=surge - low_pass,
        attrs={
            "long_name": "CoDEC surge component highpass filtered",
            "comment": f"Highpass filtering is done by removing {window_size // 24} day running mean",
            "units": "mm",
            "positive": "up",
        },
    )


def remove_slr_from_gtsm_tide_slr(tide_slr, conf):
    """
    Remove sea-level rise from the tide+slr time series output by GTSM.
    Parameters
    ----------
    tide_slr : xr.DataArray
        tide+slr time series output by GTSM

    conf: argparse.Namespace
        experiment specific configurations

    Returns
    -------
    tide : xr.DataArray
        tide without sea-level rise
    """

    tot_sea_lev_file = conf.preprocessed_input / Path(
        "totalSeaLevel_extrapolated_dis.nc"
    )
    tot_sea_lev = xr.open_dataset(tot_sea_lev_file)["TotalRelativeSeaLevel"]
    nearest_series = tot_sea_lev.sel(
        lat=tide_slr.lat, lon=tide_slr.lon % 360, method="nearest"
    )
    tide = (
        tide_slr.to_numpy()
        - nearest_series.sel(time=tide_slr.to_dataframe().index.year).to_numpy() * 10
    )
    # tide should have zero mean
    tide = tide - tide.mean()
    return xr.DataArray(
        dims="time",
        data=tide,
        attrs={
            "long_name": "CoDEC with only astronomical tides",
            "comment": "Is derived by removing sea level rise"
            "from the GTSM run without atmospheric forcing",
            "units": "mm",
            "positive": "up",
        },
    )


def get_climatology(timeseries, monthly_time):
    """
    removes monthly climatology of hourly time series
    Parameters
    ----------
    monthly_time: xr.DataArray
        monthly hr -- only the index is used
    timeseries: xr.DataArray
        hourly time series data
    Returns
    -------
    timeseries: xr.DataArray
        hourly time series of monthly climatology
    """
    df = timeseries.to_dataframe()
    df["month"] = df.index.month
    df = df.loc[:, [timeseries.name, "month"]]
    index = (
        monthly_time.to_dataframe()["1979-01-01 00:00":]
        .resample("1h")
        .asfreq()
        .dropna()
        .index
    )
    # the index is used to interpolate the monthly climatology from gtsm in the same way the climatology from
    # satellite altimetry is interpolated

    df["climatology"] = df.loc[index, "month"].apply(
        (lambda month: df["1993":].groupby("month").mean().loc[month].item()),
    )
    interpolated = df["climatology"].interpolate(method="cubic")
    # replace the first month of 1979 with the first month of 1980 to avoid nan values in the beginning.
    # NaN values at the beginning (and end) are caused by putting the monthly values at the center of a month.
    interpolated["1979-01"] = interpolated["1980-01"].to_numpy()
    # replace the last month of 2015 with the last month of 2014
    interpolated["2015-12"] = interpolated["2014-12"].to_numpy()
    return xr.DataArray(
        dims="time",
        data=interpolated,
        attrs={
            "long_name": "seasonal cycle of CoDEC surge",
            "comment": "Monthly climatology of the CoDEC surge component"
            "for the same years as the satellite altimetry interpolated with cubic interpolation",
            "units": "mm",
            "positive": "up",
        },
    )


def interpolate(timeseries):
    """
    Interpolates a monthly to an hourly timeseries.
    NaN values at the beginning and end of the timeseries are replaced by the
    nearest neighbour, i.e. the monthly mean. For HR data this should only affect the
    last 16 days of 2015.
    Parameters
    ----------
    timeseries: xr.DataArray
        monthly time series
    Returns
    -------
    xr.DataArray
        hourly time series
    """

    if len(timeseries) != len(timeseries.resample(time="1m").mean()):
        raise ValueError("timeseries must be a monthly timeseries")
    # interpolate the monthly data to hourly
    hourly_ts = timeseries.interp(
        time=pd.date_range("1979", "2016", freq="1h", inclusive="left").to_numpy(),
        method="cubic",
    )
    # extrapolate "2015-12-16 00:00" to "2015-12-31 23:00" with the last value which is the average value of the month
    return hourly_ts.interpolate_na(
        dim="time", method="nearest", fill_value="extrapolate"
    )


def combine_gtsm_hr(
    filenames, conf, vlm_oelsmann_ds, vlm_hammond, sed_frederikse_2020, index
):
    """
    Computational stack to create a combined sea-level time series from the gtsm and the hr datasets for one station.

    Parameters
    ----------
    index : int
        index of the current job
    filenames : [str]
        contains a sorted list of all file names
    conf : argparse.Namespace
        experiment specific configurations
    vlm_oelsmann_ds : xarray.Dataset
        dataset with nearest neighbour in the oelsmann vlm dataset for each station
    vlm_hammond : xarray.Dataset
        dataset with nearest neighbour in the hammond vlm dataset for each station
    Returns
    -------
    hourly_data, monthly_data, parameters : xr.Dataset, xr.Dataset, xr.Dataset
        all hourly outputs, all monthly outputs and all parameters (single values for each station)
    """
    filename = filenames[index]

    save_dir_monthly = conf.output_dir / "monthly_time_series"
    save_dir_monthly.mkdir(parents=True, exist_ok=True)
    outfile_monthly = save_dir_monthly / f"{filename}.nc"

    save_dir_hourly = conf.output_dir / "hourly_time_series"
    save_dir_hourly.mkdir(parents=True, exist_ok=True)
    outfile_hourly = save_dir_hourly / f"{filename}.nc"

    save_dir_params = conf.output_dir / "parameters"
    save_dir_params.mkdir(parents=True, exist_ok=True)
    outfile_params = save_dir_params / f"{filename}.nc"

    if (
        outfile_hourly.is_file()
        and outfile_monthly.is_file()
        and outfile_params.is_file()
        and not conf.overwrite
    ):
        print("found output file, skip computation")
        hourly_data = xr.load_dataset(outfile_hourly)
        monthly_data = xr.load_dataset(outfile_monthly)
        parameters = xr.load_dataset(outfile_params)
        return hourly_data, monthly_data, parameters
    else:
        # gtsm
        gtsm_tide_surge_slr = load_gtsm_output(
            conf.input_dir / "gtsm_tide_surge_slr" / f"{filename}.nc", conf
        )
        gtsm_tide_slr = load_gtsm_output(
            conf.input_dir / "gtsm_tide_slr" / f"{filename}.nc", conf
        )
        monthly_data = xr.load_dataset(conf.input_dir / "hr_monthly" / f"{filename}.nc")
        standard_attrs = {"units": "mm", "positive": "up"}
        monthly_data["is_dry"] = gtsm_tide_surge_slr.is_dry
        monthly_data["gia"] = xr.DataArray(
            dims="time",
            data=monthly_data["hr_no_season"] - monthly_data["hr_no_gia_no_season"],
            attrs={
                "long_name": "Glacial Isostatic Adjustment",
                "comment": "It is the RSL sign convention, i.e. negative means uplift, positive is subsidence",
                "references": "https://doi.org/10.1038/s41558-019-0531-8",
                **standard_attrs,
            },
        )

        hourly_data = xr.Dataset(
            data_vars={
                "gtsm_tide_surge_slr": xr.DataArray(
                    dims="time",
                    data=gtsm_tide_surge_slr.waterlevel.loc[:"2015"].to_numpy(),
                    attrs={
                        "long_name": "Coastal Dataset for the Evaluation of Climate Impact",
                        "comment": "Coastal waterlevels modelled with the global tide an surge model (GTSM) "
                        "including surges, astronomical tides and long term sea level rise",
                        "references": "https://doi.org/10.3389/fmars.2020.00263",
                        **standard_attrs,
                    },
                ),
                "gtsm_tide_slr": xr.DataArray(
                    dims="time",
                    data=gtsm_tide_slr.waterlevel.loc[:"2015"].to_numpy(),
                    attrs={
                        "long_name": "CoDEC without surges",
                        "comment": "Global tide an surge model run without atmospheric forcing",
                        **standard_attrs,
                    },
                ),
                "gtsm_surge": xr.DataArray(
                    dims="time",
                    data=gtsm_tide_surge_slr.waterlevel.loc[:"2015"].to_numpy()
                    - gtsm_tide_slr.waterlevel.loc[:"2015"].to_numpy(),
                    attrs={
                        "long_name": "CoDEC only surges",
                        "comment": "Derived by removing a global tide an surge model"
                        " run without atmospheric forcing from the full "
                        "CoDEC data",
                        **standard_attrs,
                    },
                ),
                "gia": xr.DataArray(
                    dims="time",
                    data=interpolate(monthly_data["gia"]),
                    attrs={
                        "long_name": "Glacial isostatic adjustment",
                        "comment": "Interpolated from the monthly time series with cubic interpolation",
                        **standard_attrs,
                    },
                ),
                "seasonal_cycle_altimetry": xr.DataArray(
                    dims="time",
                    data=interpolate(monthly_data["seasonal_cycle"]),
                    attrs={
                        "long_name": "Seasonal cycle from satellite altimetry",
                        "comment": "Monthly seasonal cycle from satellite altimetry "
                        "interpolated with cubic interpolation",
                        **standard_attrs,
                    },
                ),
                "is_dry": gtsm_tide_surge_slr.is_dry,
            },
            coords={
                "time": xr.DataArray(
                    dims="time",
                    data=gtsm_tide_surge_slr["index"].loc[:"2015"].to_numpy(),
                    attrs=gtsm_tide_surge_slr.attrs,
                ),
                "lat": monthly_data.lat,
                "lon": monthly_data.lon,
                "station_name": gtsm_tide_surge_slr["station_name"],
            },
            attrs={
                "Convention": "CF-1.9",
                "featureType": "timeSeries",
                "title": "Hourly Coastal water levels with Counterfactuals (HCC)",
                # "source": "", # todo
                # "reference": "",  # todo
            },
        )

        hourly_data["gtsm_tide"] = remove_slr_from_gtsm_tide_slr(
            tide_slr=hourly_data["gtsm_tide_slr"], conf=conf,
        )
        hourly_data["seasonal_cycle_gtsm"] = get_climatology(
            hourly_data["gtsm_surge"], monthly_time=monthly_data.time
        )
        hourly_data["gtsm_surge_no_season"] = xr.DataArray(
            dims="time",
            data=hourly_data["gtsm_surge"] - hourly_data["seasonal_cycle_gtsm"],
            attrs={
                "long_name": "CoDEC surge component without seasonal cycle",
                **standard_attrs,
            },
        )
        window_size = conf.cutoff_frequency * 30 * 24
        hourly_data["gtsm_surge_highpass"] = highpass_filter(
            surge=hourly_data["gtsm_surge_no_season"], window_size=window_size,
        )
        # crustal_pd
        hourly_index = hourly_data.to_dataframe().index
        days_per_year = 365 + hourly_index.is_leap_year
        x_hourly = (
            hourly_index.year
            + ((hourly_index.day_of_year - 1) + hourly_index.hour / 24) / days_per_year
        ).to_numpy()
        crustal_pd_intercept, crustal_pd_slope, x_monthly = lin_regr(
            monthly_data["Crustal_PD"]
        )
        hourly_data["crustal_pd_linear"] = xr.DataArray(
            dims="time",
            data=(x_hourly * crustal_pd_slope) - x_hourly[-1] * crustal_pd_slope,
            attrs={
                "long_name": "Linear approximation of crustal deformation from present day ice melt",
                "comment": "Linear fit to Crustal_PD extrapolated back to the year 1900",
                "units": "mm",
                "positive": "down",
            },
        )
        monthly_data["crustal_pd_linear"] = xr.DataArray(
            dims="time",
            data=(x_monthly * crustal_pd_slope) - x_monthly[-1] * crustal_pd_slope,
            attrs={
                "long_name": "Linear approximation of crustal deformation from present day ice melt",
                "comment": "Linear fit to Crustal_PD extrapolated back to the year 1900",
                "units": "mm",
                "positive": "down",
            },
        )

        # some assertion to make sure we do not get mixed up with the file ids
        # get file_idx to load the correct vlm value
        with xr.load_dataset(conf.input_dir / "coor_coastal.nc") as coor_coastal:
            file_idx = np.where(
                coor_coastal["station_name"] == gtsm_tide_surge_slr["station_name"]
            )
            assert coor_coastal["station_y_coordinate"][file_idx].item() == approx(
                gtsm_tide_surge_slr["latitude"].item()
            )
            assert coor_coastal["station_x_coordinate"][file_idx].item() == approx(
                gtsm_tide_surge_slr["longitude"].item()
            )
        # hammond vlm

        assert vlm_hammond["lat"][file_idx].item() == approx(
            gtsm_tide_surge_slr["latitude"]
        )
        assert vlm_hammond["lon"][file_idx].item() == approx(
            gtsm_tide_surge_slr["longitude"]
        )
        vlm_rate = vlm_hammond["vlm_rate"][file_idx].item()
        hourly_data["vlm_linear_hammond"] = xr.DataArray(
            dims="time",
            data=(x_hourly * vlm_rate) - x_hourly[-1] * vlm_rate,
            attrs={
                "long_name": "linear fit to vertical land motion from Hammond et al. 2021",
                "comment": "It does not follow the RSL sign convention, "
                "i.e. negative means subsidence, positive is uplift",
                "references": "https://doi.org/10.1029/2021JB022355",
                "positive": "down",
                "units": "mm",
            },
        )
        monthly_data["vlm_linear_hammond"] = xr.DataArray(
            dims="time",
            data=(x_monthly * vlm_rate) - x_monthly[-1] * vlm_rate,
            attrs={
                "long_name": "linear fit to vertical land motion from gps stations",
                "comment": "It does not follow the RSL sign convention, "
                "i.e. negative means subsidence, positive is uplift",
                "references": "https://doi.org/10.1029/2021JB022355",
                "units": "mm",
                "positive": "down",
            },
        )

        # oelsmann vlm
        vlm_oelsmann = vlm_oelsmann_ds.loc[
            {"station_name": monthly_data.station_name, "time": slice("1900", "2015")}
        ]["vlm"]

        monthly_data["vlm_oelsmann"] = xr.DataArray(
            dims="time",
            data=vlm_oelsmann - vlm_oelsmann[-1],
            attrs={
                "long_name": "Vertical land motion from Oelsmann et al. 2022",
                "comment": "It does not follow the RSL sign convention, "
                "i.e. negative means subsidence, positive is uplift",
                "references": "https://doi.org/10.5194/egusphere-egu22-5281",
                "units": "mm",
                "positive": "down",
            },
        )

        # VLM due to solid earth deformation (SED) from frederikse et al. (2020)
        sed = sed_frederikse_2020.sel(
            lat=monthly_data.lat, lon=monthly_data.lon, method="nearest"
        )
        sed["time"] = pd.to_datetime(sed.time, format="%Y")

        monthly_sed = sed.resample(time="MS").asfreq()

        monthly_sed["time"] = pd.DatetimeIndex(
            monthly_sed.time
        ) + pd.offsets.DateOffset(days=15)

        monthly_sed = monthly_sed.interpolate_na(dim="time", method="linear")
        m_sed = monthly_sed.loc[{"time": slice("1900", "2015")}]
        m_sed = m_sed - m_sed[-1]

        monthly_data["sed_frederikse_2020"] = xr.DataArray(
            dims="time",
            data=m_sed,
            attrs={
                "long_name": "Glaciers, ice sheets, and terrestrial water storage Solid-Earth deformation from Frederikse et al. 2020",
                "comment": " Positive means uplift of land, negative is subsidence of land. The influence on relative sea level has the opposite sign",
                "references": "https://doi.org/10.1038/s41586-020-2591-3",
                "units": "mm",
                "positive": "down",
            },
        )

        # derive geocentric water levels from HR by excuding GIA and SED effects
        hr_abs = (
            monthly_data["hr_no_gia_no_season"]
            + monthly_data["sed_frederikse_2020"]
            + monthly_data.seasonal_cycle
        )
        msl_hr_abs = hr_abs.loc[{"time": slice("1993", "2012")}].mean(dim="time")
        hr_aviso_offset = msl_hr_abs - monthly_data["mdt_global"]

        # rel and abs sl
        monthly_data[f"hr_absolute_no_season"] = xr.DataArray(
            dims="time",
            data=monthly_data["hr_no_gia_no_season"]
            + monthly_data["crustal_pd_linear"]
            - hr_aviso_offset,
            attrs={
                "long_name": "Absolute coastal water levels without seasonal cycle",
                "comment": "height of the sea surface above reference geoid",
                **standard_attrs,
            },
        )

        # relative sea level
        monthly_data[f"hr_relative_no_season"] = xr.DataArray(
            dims="time",
            data=monthly_data[f"hr_absolute_no_season"] - monthly_data["vlm_oelsmann"],
            attrs={
                "long_name": "Relative coastal water levels including vertical land motion without seasonal cycle",
                "comment": "height of the sea surface relative to the height of the coast",
                **standard_attrs,
            },
        )

        parameters = xr.Dataset(
            data_vars={
                "vlm_rate_hammond": xr.DataArray(
                    data=vlm_rate,
                    attrs={
                        "long_name": "Rate of vertical land motion from Hammond et al. (2021)",
                        "comment": "Linear rate of vertical land motion estimated from GPS measurements"
                        "Negative means subsidence of land, positive is uplift of land."
                        "The influence on relative sea level has opposite signs.",
                        "references": "https://doi.org/10.1029/2021JB022355",
                        "positive": "down",
                        "units": "mm.yr-1",
                    },
                ),
                "crustal_pd_slope": xr.DataArray(
                    data=crustal_pd_slope,
                    attrs={
                        "long_name": "Slope of linear vertical land motion from present day ice melt",
                        "comment": "Negative means subsidence of land, positive is uplift of land."
                        "The influence on relative sea level has opposite signs.",
                        "positive": "down",
                        "units": "mm.yr-1",
                    },
                ),
                "crustal_pd_intercept": xr.DataArray(
                    data=crustal_pd_intercept,
                    attrs={
                        "long_name": "Intercept of linear vertical land motion from present day ice melt",
                        "comment": "Negative means subsidence of land, positive is uplift of land."
                        "The influence on relative sea level has opposite signs.",
                        "units": "mm",
                        "positive": "down",
                    },
                ),
                "is_dry": gtsm_tide_surge_slr.is_dry,
            },
            coords={
                "lat": hourly_data.lat,
                "lon": hourly_data.lon,
                "station_name": hourly_data.station_name,
                "file_name": filename,
            },
            attrs={
                "Convention": "CF-1.9",
                "featureType": "point",
                "title": "Parameters for the HCC dataset",
                # "source": "", # todo
                # "reference": "",  # todo
            },
        )

        for sl_type in ["absolute", "relative"]:
            hourly_data[f"hr_{sl_type}_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_no_season"]),
                attrs={
                    "long_name": "Interpolated "
                    + monthly_data[f"hr_{sl_type}_no_season"].attrs["long_name"],
                    "comment": monthly_data[f"hr_{sl_type}_no_season"].attrs["comment"],
                    **standard_attrs,
                },
            )

            regression_ds = RegressionDataset(monthly_data[f"hr_{sl_type}_no_season"])
            x = pd.DataFrame(
                {
                    "t": regression_ds.t.to_numpy(),
                    "t_squared": regression_ds.t_square.to_numpy(),
                },
                index=regression_ds.t.time,
            )
            y = regression_ds.normalized_ts.to_dataframe()[
                regression_ds.normalized_ts.name
            ]
            mod = sm.OLS(y.resample("1a").mean(), x.resample("1a").mean(),)
            res = mod.fit()

            monthly_data[f"hr_{sl_type}_trend_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(res.predict(x)),
                attrs={
                    "long_name": f"Quadratic trend of {sl_type} coastal water levels",
                    "comment": f"Computed by fitting a quadratic function to the {sl_type} coastal water levels."
                    "The intercept is fixed at the average sea level of 1900-1906",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_trend_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_trend_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_trend_no_season"].attrs,
            )
            monthly_data[f"hr_{sl_type}_trend_l_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(res.get_prediction(x).conf_int()[:, 0]),
                attrs={
                    "long_name": f"lower percentile of the 95 percent confidence interval "
                    f"of the quadratic trend of {sl_type} coastal water levels",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_trend_l_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_trend_l_pctl_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_trend_l_pctl_no_season"].attrs,
            )
            monthly_data[f"hr_{sl_type}_trend_u_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(res.get_prediction(x).conf_int()[:, 1]),
                attrs={
                    "long_name": f"upper percentile of the 95 percent confidence interval "
                    f"of the quadratic trend of {sl_type} coastal water levels",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_trend_u_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_trend_u_pctl_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_trend_u_pctl_no_season"].attrs,
            )
            monthly_data[f"hr_{sl_type}_cfact_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(y - res.predict(x)),
                attrs={
                    "long_name": f"{sl_type} coastal water levels without quadratic trend",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_cfact_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_cfact_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_cfact_no_season"].attrs,
            )
            # subtracting the upper percentile of the trend gives
            # the lower percentile of the counterfactual and vice versa
            monthly_data[f"hr_{sl_type}_cfact_u_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(y - res.get_prediction(x).conf_int()[:, 0]),
                attrs={
                    "long_name": f"Upper percentile of a 95 percent confidence interval of"
                    f"{sl_type} coastal water levels without quadratic trend",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_cfact_u_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_cfact_u_pctl_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_cfact_u_pctl_no_season"].attrs,
            )
            monthly_data[f"hr_{sl_type}_cfact_l_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=regression_ds.rescale(y - res.get_prediction(x).conf_int()[:, 1]),
                attrs={
                    "long_name": f"Lower percentile of a 95 percent confidence interval of"
                    f"{sl_type} coastal water levels without quadratic trend",
                    **standard_attrs,
                },
            )
            hourly_data[f"hr_{sl_type}_cfact_l_pctl_no_season"] = xr.DataArray(
                dims="time",
                data=interpolate(monthly_data[f"hr_{sl_type}_cfact_l_pctl_no_season"]),
                attrs=monthly_data[f"hr_{sl_type}_cfact_l_pctl_no_season"].attrs,
            )
            for c, data, unit in [
                ("b", res.params["t"], "mm.yr-1"),
                ("b_l_pctl", res.conf_int().transpose()["t"][0], "mm.yr-1"),
                ("b_u_pctl", res.conf_int().transpose()["t"][1], "mm.yr-1"),
                ("a", res.params["t_squared"], "mm.yr-2"),
                ("a_l_pctl", res.conf_int().transpose()["t_squared"][0], "mm.yr-2"),
                ("a_u_pctl", res.conf_int().transpose()["t_squared"][1], "mm.yr-2"),
            ]:
                parameters[f"{sl_type}_sl_{c}"] = xr.DataArray(
                    data=regression_ds.rescale(data),
                    attrs={
                        "long_name": f"Parameter {c} for {sl_type} sea level",
                        "comment": f"The quadratic trend is calculated with the formular at^2 + bt + c = sl_trend . "
                        f"c is fixed to be 0. The parameters are fitted on sea level relative to the mean sea level "
                        f"from 1900-1906"
                        f"t [years] is a floating point starting at 0",
                        "units": unit,
                    },
                )
            parameters[f"{sl_type}_sl_rsquared"] = xr.DataArray(
                data=res.rsquared,
                attrs={
                    "long_name": f"Coefficient of determination for {sl_type} sea level",
                    "units": "1",
                },
            )

            for cfact, attrs in [
                (
                    "",
                    {
                        "long_name": f"{sl_type} hourly coastal water levels with long-term trends",
                        "comment": f"{sl_type} sea level as a combination of "
                        f"highpass-filtered CoDEC surges, CoDEC astronomical tides and lowpass-filtered hr sea level"
                        f"with the seasonal cycle added from satellite altimetry data",
                        **standard_attrs,
                    },
                ),
                (
                    "_cfact",
                    {
                        "long_name": f"Counterfactual {sl_type} hourly coastal water levels with long-term trends ",
                        "comment": f"{sl_type} HCC without the quadratic trend of sea level since 1900. "
                        f"This represents a baseline without {sl_type} sea level rise since 1900",
                        **standard_attrs,
                    },
                ),
                (
                    "_cfact_u_pctl",
                    {
                        "long_name": f"Upper percentile of counterfactual {sl_type} HCC",
                        "comment": f"Upper percentile of a 95 percent confidence interval",
                        **standard_attrs,
                    },
                ),
                (
                    "_cfact_l_pctl",
                    {
                        "long_name": f"Lower percentile of counterfactual {sl_type} HCC",
                        "comment": f"Lower percentile of a 95 percent confidence interval",
                        **standard_attrs,
                    },
                ),
            ]:
                hr_no_season_lowpass_filtered = (
                    hourly_data[f"hr_{sl_type}{cfact}_no_season"]
                    if conf.cutoff_frequency == 1
                    else hourly_data[f"hr_{sl_type}{cfact}_no_season"]
                    .rolling(
                        time=window_size, center=True, min_periods=window_size // 2
                    )
                    .mean()
                )
                hourly_data[f"hcc_{sl_type}{cfact}"] = xr.DataArray(
                    dims="time",
                    data=(
                        hr_no_season_lowpass_filtered
                        + hourly_data["gtsm_tide"]
                        + hourly_data["gtsm_surge_highpass"]
                        + hourly_data["seasonal_cycle_altimetry"]
                    ),
                    attrs=attrs,
                )

        monthly_data.to_netcdf(outfile_monthly, format="NETCDF4_CLASSIC")
        hourly_data.to_netcdf(outfile_hourly, format="NETCDF4_CLASSIC")

        # df_params
        # add means over the first 6 years and the last 6 years for parameters
        # here use pandas for slicing time indices
        df_monthly = monthly_data.drop_vars("time_bounds").to_dataframe()
        df_hourly = hourly_data.to_dataframe()
        for data_array in monthly_data:
            if "hr" in data_array:
                for idx in [
                    df_monthly.first("6a").index,
                    df_monthly[
                        df_hourly.first("6a").index[0] : df_hourly.first("6a").index[-1]
                    ].index,
                    df_monthly.last("6a").index,
                ]:
                    parameters[
                        f"{data_array}_{idx[0].year}-{idx[-1].year}"
                    ] = xr.DataArray(
                        data=df_monthly.loc[idx, data_array].mean(),
                        attrs={
                            "long_name": "Mean sea level from {idx[0].year} to {idx[-1].year for }"
                            + monthly_data[data_array].long_name,
                            "units": "mm",
                        },
                    )

        # add means over the first 6 years and the last 6 years for df_hourly
        for data_array in hourly_data:
            if "hcc" in data_array:
                for idx in [df_hourly.first("6a").index, df_hourly.last("6a").index]:
                    parameters[
                        f"{data_array}_{idx[0].year}-{idx[-1].year}"
                    ] = xr.DataArray(
                        data=df_hourly.loc[idx, data_array].mean(),
                        attrs={
                            "long_name": "Mean sea level from {idx[0].year} to {idx[-1].year for }"
                            + hourly_data[data_array].long_name,
                            "units": "mm",
                        },
                    )

            idx = df_hourly.first("6a").index
            if "trend" in data_array:
                parameters[f"{data_array}_{idx[0].year}-{idx[-1].year}"] = xr.DataArray(
                    data=df_monthly.loc[idx[0] : idx[-1], data_array].mean(),
                    attrs={
                        "long_name": "Mean sea level from {idx[0].year} to {idx[-1].year for }"
                        + monthly_data[data_array].long_name,
                        "units": "mm",
                    },
                )

        parameters.to_netcdf(outfile_params, format="NETCDF4_CLASSIC")
        print("RAM Used (GB):", psutil.virtual_memory()[3] / 1000000000)
        return hourly_data, monthly_data, parameters
