"""project configs via argparse and static global variables"""
import argparse
import os
from pathlib import Path
import logging


def save_args(args, file_name):
    """
    Saves the command line arguments to a file.
    The experiment can be re-run by passing '@file' as commandline argument.
    Parameters
    ----------
    args : dict
        command line arguments
    file_name : str, pathlib.Path
        path to args.txt file
    Returns
    -------

    """
    with open(file_name, "w") as opt_file:
        for k, v in sorted(args.items()):
            if type(v) == bool:
                if v:
                    opt_file.write(f"--{str(k)}\n")
            elif type(v) == list:
                opt_file.write(f"--{str(k)}\n")
                for element in v:
                    opt_file.write(f"{str(element)}\n")
            else:
                opt_file.write(f"--{str(k)}\n{str(v)}\n")


class ConfigParser:
    """
    Parses the configuration settings for one experiment with argparse
    """

    def __init__(self):
        """
        initializes a parser with arguments for all settings that may change between experiments.
        """
        self.parser = argparse.ArgumentParser(fromfile_prefix_chars="@")
        # required arguments
        self.parser.add_argument("--name", type=str, help="Name of the experiment")
        # optional arguments
        self.parser.add_argument(
            "--input_dir",
            type=str,
            default=None,
            help="Default input_dir BASE_DIR/input",
        )
        self.parser.add_argument(
            "--output_base",
            type=str,
            default=None,
            help="Default output_base BASE_DIR/output",
        )
        self.parser.add_argument(
            "--output_dir", type=str, default=None, help="specifies dir for output"
        )
        self.parser.add_argument(
            "--overwrite",
            action="store_true",
            help="If specified, saved precomputed data will be overwritten",
        )
        self.parser.add_argument(
            "--timeout",
            type=int,
            default=2000,
            help="maximum time in seconds for each computation block to run",
        )
        self.parser.add_argument(
            "--n_test_stations",
            type=int,
            default=6,
            help="number of test_stations for unit tests",
        )
        self.parser.add_argument(
            "--preprocessed_input",
            type=str,
            default=None,
            help="specifies the location of preprocessed input data. "
            "Default is BASE_DIR/input/preprocessed_input",
        )
        self.parser.add_argument(
            "--id_lon_lat_file",
            type=str,
            default=None,
            help="specifies path of gtsm_id_lon_lat",
        )
        self.parser.add_argument(
            "--gesla_2",
            type=str,
            default="unique",
            help="specifies gesla_2 dataset. Use 'public' for data that is downloaded from the gesla2 webpage"
            "and 'unique' for preprocessed data from Thomas Wahl that comes as .mat files",
        )

        self.parser.add_argument(
            "--validation_time_scales",
            nargs="+",
            help="time scales for which validation measures are specified. Multiple options are allowed",
            default="1d",
        )
        self.parser.add_argument(
            "--cutoff_frequency",
            type=int,
            default=3,
            help="variability with frequencies below cutoff_frequency*30d comes from gtsm. "
            "Frequencies above that come from HR",
        )

    def parse(self, args=None):
        """
        parses the arguments and sets up an experiment
        Parameters
        ----------
        args : [str] default = None
            if None arguments are taken from the command line. If the main file is run from a script,
            arguments can be passed as a list of strings. See also the argparse documentation
        Returns
        -------

        """
        if args:
            conf = self.parser.parse_args(args)
        else:
            conf = self.parser.parse_args()
        root_dir = Path(os.path.dirname(os.path.abspath(__file__))).parent
        if conf.output_base is None:
            conf.output_base = root_dir / Path("output")
        else:
            conf.output_base = Path(conf.output_base)
        if conf.input_dir is None:
            conf.input_dir = root_dir / Path("input")
        else:
            conf.input_dir = Path(conf.input_dir)
        if conf.preprocessed_input is None:
            conf.preprocessed_input = root_dir / "input" / "preprocessed_input"
        else:
            conf.preprocessed_input = Path(conf.preprocessed_input)
        conf.preprocessed_input.mkdir(parents=True, exist_ok=True)
        if conf.output_dir is None:
            conf.output_dir = conf.output_base / Path(conf.name)
        else:
            conf.output_dir = Path(conf.output_dir)
        conf.output_dir.mkdir(parents=True, exist_ok=True)
        # id_lon_lat contains all station ids, corresponding coordinates and bedlevel.
        # It is used to ensure the correct datasets are put together.
        if conf.id_lon_lat_file is None:
            conf.id_lon_lat_file = (
                Path(root_dir) / "input" / "03_ERA5_id_lon_lat_bedlev.xyzn"
            )
        else:
            conf.id_lon_lat_file = Path(conf.id_lon_lat_file)

        # default ore input of a single time scale are not parsed as a list
        if type(conf.validation_time_scales) == str:
            conf.validation_time_scales = [conf.validation_time_scales]
        return conf


def logging_setup(output_dir, overwrite=False):
    """
    Set up an error log. This log collects errors from different workers in a single file.
    Parameters
    ----------
    output_dir : Path
        logfile path
    overwrite : Boolean
        if True overwrite existing error.log
    Returns
    -------
    logger : logging.Logger
    """
    if overwrite:
        filemode = "w"
    else:
        filemode = "a"
    logging.basicConfig(
        filename=str(output_dir / "error.log"),
        filemode=filemode,
        level=logging.WARNING,
        format="%(asctime)s %(levelname)s %(name)s %(message)s",
    )
    logger = logging.getLogger(__name__)
    return logger
