"""helper scripts for data loading"""
from cf_units import Unit
from gslcomp.config import logging_setup
import netCDF4 as nc
import numpy as np
import cftime
import xarray as xr


def get_input_filenames(input_dir):
    """
    gets a list of all filenames in input_dir/gtsm_tide_surge_slr

    Parameters
    ----------
    input_dir: pathlib.Path
        input directory (usually from conf)

    Returns
    -------
    filenames: [str]
        sorted list of filenames without file ending
    """
    gtsm_tide_surge_slr_dir = input_dir / "gtsm_tide_surge_slr"
    assert gtsm_tide_surge_slr_dir.is_dir()
    filenames = [p.stem for p in gtsm_tide_surge_slr_dir.glob("gtsm_station*.nc")]
    return sorted(filenames)


def load_gtsm_output(file, conf):
    """
    Loading time series that are output by the gtsm model.

    Parameters
    ----------
    file : pathlib.Path
        path to netcdf file
    conf : argparse.Namespace
        experiment specific configuration

    Returns
    -------
    dataset : xr.Dataset
    """
    gtsm_dataset = xr.load_dataset(file)
    logger = logging_setup(conf.output_dir)
    is_dry = False
    dry = gtsm_dataset.sel(
        {
            "index": gtsm_dataset.waterlevel
            <= Unit(gtsm_dataset.bedlevel.units).convert(
                gtsm_dataset.bedlevel, gtsm_dataset.waterlevel.units
            )
        }
    )
    # bedlevel is in meters and waterlevel in mm
    if dry.waterlevel.size == gtsm_dataset.waterlevel.size:
        logger.warning(
            f"\n\t\tAll values are <= bedlevel of {gtsm_dataset.bedlevel} meters \n"
            f"\t\tat the station {file.stem} at\n"
            f"\t\tlat = {gtsm_dataset.latitude.item()}\n"
            f"\t\tlon = {gtsm_dataset.longitude.item}"
        )
        is_dry = True
    elif dry.waterlevel.size > 0:
        logger.warning(
            f"\n\t\tFound {dry.waterlevel.size} values <= bedlevel of {gtsm_dataset.bedlevel} meters \n"
            f"\t\tat the station {file.stem} at\n"
            f"\t\tlat = {gtsm_dataset.latitude.item()}\n"
            f"\t\tlon = {gtsm_dataset.longitude.item}"
        )
        is_dry = True

    return xr.Dataset(
        {
            **{
                var: gtsm_dataset[var]
                if var != "waterlevel"
                else gtsm_dataset[var].where(
                    gtsm_dataset[var]
                    > Unit(gtsm_dataset.bedlevel.units).convert(
                        gtsm_dataset.bedlevel, gtsm_dataset.waterlevel.units
                    )
                )
                for var in gtsm_dataset
            },
            "is_dry": xr.DataArray(
                data=is_dry,
                attrs={
                    "long_name": "Waterlevel falls dry at this location",
                    "flag_values": [False, True],
                    "flag_meanings": "waterlevel_greater_bedlevel waterlevel_smaller_or_equal_bedlevel",
                },
            ),
        }
    )
