import numpy as np
import pandas as pd
import xarray as xr
from gslcomp.config import ConfigParser
from gslcomp.utils import get_coastline, get_reg_mask


def get_extremes(da):
    extremes_list = []
    for sn, da_station in da.groupby("station_name"):
        da_valid = da_station.where(
            ~da_station["observed_surge_level_anomaly"].isnull(), drop=True
        )
        da_sorted = da_valid.sortby(
            da_valid["observed_surge_level_anomaly"].sel(station_name=sn)
        )
        n_extreme_values = len(da_sorted.time) // 100
        extremes_list.append(da_sorted.tail(n_extreme_values))
    extremes = xr.concat(extremes_list, dim="station_name")
    return extremes.rename(
        {
            f"{dataset}_surge_level_anomaly": f"{dataset}_99pctl_surge_level_anomaly"
            for dataset in ["observed", "hcc", "hcc_counterfactual"]
        }
    )


def main(conf):
    start_year = 2011
    end_year = 2015
    extreme_value_interval = slice(f"{start_year}-01-01", f"{end_year}-12-31")

    reg_mask_xr, reg_mask = get_reg_mask(conf.input_dir / "regionMap_Thompson_inv.mat")
    with xr.open_dataset(conf.output_dir / "val_ds_extremes.nc") as ds:
        station_name = ds.station_name
        val_nc = (
            ds.assign_coords(coastline=get_coastline(station_name, reg_mask))
            .sortby(["lat"], ascending=False)
            .sortby(["coastline"])
            .load()
        )
    val_nc["hcc_surge"] = val_nc["hcc"] - val_nc["codec_tide"]
    val_nc["hcc_cfact_surge"] = val_nc["hcc_cfact"] - val_nc["codec_tide"]

    pred_tides = xr.load_dataset(conf.output_dir / "predicted_tides.nc")
    val_nc["tg_surge"] = val_nc["w_hour"] - pred_tides["pred_tides"]
    val_nc["observed_surge_level_anomaly"] = val_nc["tg_surge"] - val_nc[
        "tg_surge"
    ].loc[{"time": extreme_value_interval}].mean(dim="time")
    hcc_surge_mean = (
        val_nc["hcc_surge"]
        .where(~val_nc["tg_surge"].isnull())
        .loc[{"time": extreme_value_interval}]
        .mean(dim="time")
    )
    val_nc["hcc_surge_level_anomaly"] = val_nc["hcc_surge"] - hcc_surge_mean
    val_nc["hcc_counterfactual_surge_level_anomaly"] = (
        val_nc["hcc_cfact_surge"] - hcc_surge_mean
    )
    extremes = get_extremes(
        val_nc[
            [
                f"{dataset}_surge_level_anomaly"
                for dataset in ["observed", "hcc", "hcc_counterfactual"]
            ]
        ]
        .loc[{"time": extreme_value_interval}]
        .resample(time="1D")
        .max()
    )

    # add coastline
    extremes["coastline"] = get_coastline(extremes, reg_mask)
    extremes = extremes.sortby(["lat"], ascending=False).sortby(["coastline"])
    extremes["observed_99pctl_surge_level_anomaly"].attrs[
        "long_name"
    ] = f"99th percentile of observed daily maximum surge level anomalies from {start_year}-{end_year}"
    extremes["hcc_99pctl_surge_level_anomaly"].attrs[
        "long_name"
    ] = "HCC surge level anomalies at the same time steps as observed_99pctl_surge_level_anomaly"
    extremes["hcc_counterfactual_99pctl_surge_level_anomaly"].attrs[
        "long_name"
    ] = "HCC counterfactual surge level anomalies at the same time steps as observed_99pctl_surge_level_anomaly"
    extremes["hcc_water_level_monthly"] = val_nc["hcc_long"]
    extremes["hcc_geocentric_water_level_monthly"] = val_nc["geocentric_hcc_long"]

    extremes.to_netcdf(
        conf.output_dir / f"hcc_gesla3_99pctl_surge_{start_year}_{end_year}.nc"
    )


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
