from pathlib import Path

import numpy as np
import pandas as pd
import statsmodels.api as sm
import xarray as xr
from gslcomp.config import ConfigParser
from gslcomp.utils import (get_coastline, get_index_of_nearest_locations,
                           get_reg_mask)

FRAC_VALID_DATA = 0.3
TIME_INTERVAL = slice("1993", "2012")
MIN_YEARS_OBS_PSMSL = 20


def get_rate(da):
    if len(da.dropna(dim="time")) > 3:
        x_tg = da.time.dt.year.to_numpy()
        y_tg = da.to_numpy()
        X = np.expand_dims(x_tg, axis=1)
        X = sm.add_constant(X)
        model = sm.OLS(y_tg, X, missing="drop")
        results = model.fit()
        try:
            obs = results.params[1]
            obs_lower = results.conf_int()[1, 0]
            obs_upper = results.conf_int()[1, 1]
        except IndexError as e:
            print(e)
    else:
        obs = np.NaN
        obs_lower = np.NaN
        obs_upper = np.NaN
    return obs, obs_lower, obs_upper


def main(conf):
    # Settings to select tide gauges

    # Tide Gauges are vertically aligned to have the same mean values
    # from 1993 - 2015 as HCC
    # (or for the available years of observations in this period)
    # To ensure enough data in this alignment period
    # we select only tide gauges with at least 30 percent coverage

    with xr.open_dataset(conf.output_dir / "val_ds_psmsl.nc") as ds:
        ds_sel = ds[["psmsl"]].load()

    selection_psmsl = ds_sel["psmsl"].where(
        (
            ds_sel.psmsl.loc[{"monthly_time": TIME_INTERVAL}].count(dim="monthly_time")
            >= FRAC_VALID_DATA
            * ds_sel.monthly_time.loc[{"monthly_time": TIME_INTERVAL}].count(
                dim="monthly_time"
            )
        )
        & (ds_sel.psmsl.count(dim="monthly_time") >= 12 * MIN_YEARS_OBS_PSMSL),
        drop=True,
    )

    print(f"There are {len(selection_psmsl.station_name)} selected stations in PSMSL")
    print(
        f"with at least {FRAC_VALID_DATA * 100 :.1f} percent of valid",
        f"observations from {TIME_INTERVAL.start} to {TIME_INTERVAL.stop} ",
    )

    reg_mask_xr, reg_mask = get_reg_mask(conf.input_dir / "regionMap_Thompson_inv.mat")

    selection_psmsl = selection_psmsl.assign_coords(
        coastline=get_coastline(selection_psmsl.station_name, reg_mask)
    )

    sorted_station_name = (
        selection_psmsl.station_name.drop_duplicates(dim="station_name")
        .sortby(["lat"], ascending=False)
        .sortby(["coastline"])
    )

    with xr.open_dataset(conf.output_dir / "val_ds_psmsl.nc") as ds:
        ds_sel = ds[
            [
                "hcc_long",
                "geocentric_hcc_long",
                "hcc_cfact_long",
                "geocentric_hcc_cfact_long",
                "psmsl",
                "hr",
            ]
        ]
        val_nc = ds_sel.load().sel(station_name=sorted_station_name)

    alignment_offset_psmsl = (
        (val_nc["hcc_long"] - val_nc["psmsl"])
        .loc[{"monthly_time": slice("1993", "2015")}]
        .mean(dim="monthly_time")
    )

    alignment_offset_hr = (
        (val_nc["hcc_long"] - val_nc["hr"])
        .where(~val_nc["psmsl"].isnull())
        .loc[{"monthly_time": slice("1993", "2015")}]
        .mean(dim="monthly_time")
    )

    val_nc["psmsl_aligned"] = val_nc["psmsl"] + alignment_offset_psmsl
    val_nc["hr_aligned"] = val_nc["hr"] + alignment_offset_hr

    variables = ["psmsl", "hcc_long", "hr"]  # , "hr_rel_hammond"]
    slr_lists = {}
    for var in variables:
        slr_lists[f"{var}_rslr"] = []
        slr_lists[f"{var}_rslr_lower"] = []
        slr_lists[f"{var}_rslr_upper"] = []
        for sn in val_nc.station_name:
            da = val_nc.sel(station_name=sn)
            rslr, rslr_lower, rslr_upper = get_rate(
                da[var]
                .where(~da["psmsl"].isnull(), drop=True)
                .rename(monthly_time="time")
            )
            slr_lists[f"{var}_rslr"].append(rslr)
            slr_lists[f"{var}_rslr_lower"].append(rslr_lower)
            slr_lists[f"{var}_rslr_upper"].append(rslr_upper)

    slr_ds = xr.Dataset(
        data_vars={k: ("station_name", slr_lists[k]) for k in slr_lists.keys()},
        coords=val_nc.coords,
    )

    slr_ds.to_netcdf(conf.output_dir / "slr_ds_psmsl_selected.nc")

    val_ds_psmsl_selected = val_nc[
        [
            "hcc_long",
            "hr_aligned",
            "psmsl_aligned",
        ]
    ]
    val_ds_psmsl_selected = val_ds_psmsl_selected.rename(
        {
            "hcc_long": "hcc_water_level_monthly",
            "hr_aligned": "hr_aligned_water_level_monthly",
            "psmsl_aligned": "psmsl_aligned_water_level_monthly",
        }
    )
    val_ds_psmsl_selected["hcc_water_level_monthly"].attrs[
        "long_name"
    ] = "Monthly relative water level from the HCC datase"
    val_ds_psmsl_selected["hr_aligned_water_level_monthly"].attrs["long_name"] = (
        "Monthly relative water level from the HR dataset, ",
        "aligned with the HCC relative water level ",
    )
    val_ds_psmsl_selected["psmsl_aligned_water_level_monthly"].attrs["long_name"] = (
        "Monthly relative water level from the PSMSL database, ",
        "aligned with the HCC relative water level",
    )

    val_ds_psmsl_selected.to_netcdf(
        conf.output_dir / "hcc_hr_psmsl_water_level_monthly_1900_2015.nc"
    )


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
