import numpy as np
import pandas as pd
import psutil
import xarray as xr
from dask.distributed import Client
from gesla import GeslaDataset
from gslcomp.config import ConfigParser
from gslcomp.utils import get_coastline, get_reg_mask
from joblib import Parallel, delayed
from scipy.io import loadmat
from sklearn.neighbors import BallTree

# Constants
FRAC_VALID_DATA = 0.9
TIME_INTERVAL = pd.date_range("2011-01-01 00:00", "2015-12-31 23:00", freq="1H")


def calc_annual_means_for_years_with_enough_data(hourly_data, frac_valid_data):
    """Calculate annual means for valid years with enough data availability.

    Parameters
    ----------
    hourly_data : pandas.Series
        The hourly data to calculate annual means from.
    frac_valid_data : float
        The fraction of valid data points required for a year to be considered valid.

    Returns
    -------
    annual_means : pandas.Series
        The annual means for valid years.
    """
    # Calculate the total number of hours in a year
    total_hours_in_year = 365 * 24

    # Group data by year and filter out invalid years
    grouped_by_year = hourly_data.groupby(hourly_data.index.year).count()
    valid_years = grouped_by_year[
        grouped_by_year >= frac_valid_data * total_hours_in_year
    ].index

    # Calculate annual means for valid years
    annual_means = (
        hourly_data[hourly_data.index.year.isin(valid_years)].resample("YS").mean()
    )
    return annual_means


def save_station_netcdf(
    tg_f_name, gesla3, bt_gtsm, gtsm_index, output_dir, tmp_output_dir
):
    if (tmp_output_dir / f"{tg_f_name}.nc").is_file():
        return
    tg_data, tg_meta = gesla3.file_to_pandas(tg_f_name)
    tg_data_filtered = tg_data.loc[
        (tg_data["qc_flag"] == 1) & (tg_data["use_flag"] == 1), "sea_level"
    ]
    tg_data = (
        tg_data_filtered.rolling(window="1H", closed="both", center=True)
        .mean()
        .resample("1H")
        .asfreq()
    )
    if (
        tg_data[TIME_INTERVAL[0] : TIME_INTERVAL[-1]].dropna().count()
        / len(TIME_INTERVAL)
    ) < FRAC_VALID_DATA:
        return

    search_point = np.array([[tg_meta.latitude], [tg_meta.longitude]]).T
    distance, index = bt_gtsm.query(np.deg2rad(search_point))
    earth_radius = 6371000
    distance = distance.item() * earth_radius
    gtsm_f_name = gtsm_index.iloc[index.item()].filename

    hourly_data = xr.load_dataset(
        output_dir / "hourly_time_series" / f"{gtsm_f_name}.nc"
    )
    monthly_data = xr.load_dataset(
        output_dir / "monthly_time_series" / f"{gtsm_f_name}.nc"
    )

    gesla_hourly = xr.DataArray(
        dims=["time", "station_name"],
        data=(tg_data * 1000).values.reshape(-1, 1),
        coords={
            "time": xr.DataArray(
                dims="time",
                data=tg_data.index,
            ),
            "station_name": xr.DataArray(dims="station_name", data=[tg_f_name]),
        },
        attrs={
            "long_name": "Hourly water levels from GESLA3",
            "unit": "mm",
        },
    ).loc[{"time": slice("1979", "2015")}]
    if len(gesla_hourly) == 0:
        return
    amsl_data = calc_annual_means_for_years_with_enough_data(
        tg_data * 1000, frac_valid_data=10 / 11
    )

    amsl = xr.DataArray(
        dims=["annual_time", "station_name"],
        # todo create this function
        data=amsl_data.values.reshape(-1, 1),
        coords={
            "annual_time": xr.DataArray(dims="annual_time", data=amsl_data.index),
            "station_name": xr.DataArray(dims="station_name", data=[tg_f_name]),
        },
        attrs={
            "long_name": "Annual mean water levels from GESLA-2",
            "comment": "For years with enough data",
            "units": "mm",
        },
    ).loc[{"annual_time": slice("1979", "2015")}]

    hcc_long = xr.concat(
        [
            (
                monthly_data["hr_relative_no_season"] + monthly_data["seasonal_cycle"]
            ).loc[{"time": slice("1900", "1978")}],
            hourly_data["hcc_relative"].resample(time="MS", loffset="15d").mean(),
        ],
        dim="time",
    )
    geocentric_hcc_long = xr.concat(
        [
            (
                monthly_data["hr_absolute_no_season"] + monthly_data["seasonal_cycle"]
            ).loc[{"time": slice("1900", "1978")}],
            hourly_data["hcc_absolute"].resample(time="MS", loffset="15d").mean(),
        ],
        dim="time",
    )

    hcc_cfact_long = xr.concat(
        [
            (
                monthly_data["hr_relative_cfact_no_season"]
                + monthly_data["seasonal_cycle"]
            ).loc[{"time": slice("1900", "1978")}],
            hourly_data["hcc_relative_cfact"].resample(time="MS", loffset="15d").mean(),
        ],
        dim="time",
    )
    geocentric_hcc_cfact_long = xr.concat(
        [
            (
                monthly_data["hr_absolute_cfact_no_season"]
                + monthly_data["seasonal_cycle"]
            ).loc[{"time": slice("1900", "1978")}],
            hourly_data["hcc_absolute_cfact"].resample(time="MS", loffset="15d").mean(),
        ],
        dim="time",
    )
    station_data = xr.Dataset(
        data_vars={
            "hcc": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data.hcc_relative.to_numpy()],
                attrs={
                    "long_name": "Hourly coastal water levels",
                    "unit": "mm",
                },
            ),
            "hcc_cfact": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data.hcc_relative_cfact.to_numpy()],
                attrs={
                    "long_name": "Counterfactual hourly coastal water levels",
                    "unit": "mm",
                },
            ),
            "geocentric_hcc": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data.hcc_absolute.to_numpy()],
                attrs={
                    "long_name": "Geocentric hourly coastal water levels",
                    "unit": "mm",
                },
            ),
            "geocentric_hcc_cfact": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data.hcc_absolute_cfact.to_numpy()],
                attrs={
                    "long_name": "Counterfactual geocentric hourly coastal water levels",
                    "unit": "mm",
                },
            ),
            "hcc_long": xr.DataArray(
                dims=["monthly_time", "station_name"],
                data=np.c_[hcc_long.to_numpy()],
                attrs={
                    "long_name": "Monthly coastal water levels",
                    "unit": "mm",
                },
            ),
            "hcc_cfact_long": xr.DataArray(
                dims=["monthly_time", "station_name"],
                data=np.c_[hcc_cfact_long.to_numpy()],
                attrs={
                    "long_name": "Counterfactual monthly coastal water levels",
                    "unit": "mm",
                },
            ),
            "geocentric_hcc_long": xr.DataArray(
                dims=["monthly_time", "station_name"],
                data=np.c_[geocentric_hcc_long.to_numpy()],
                attrs={
                    "long_name": "Geocentric monthly coastal water levels",
                    "unit": "mm",
                },
            ),
            "geocentric_hcc_cfact_long": xr.DataArray(
                dims=["monthly_time", "station_name"],
                data=np.c_[geocentric_hcc_cfact_long.to_numpy()],
                attrs={
                    "long_name": "Counterfactual geocentric monthly coastal water levels",
                    "unit": "mm",
                },
            ),
            "hr": xr.DataArray(
                dims=["monthly_time", "station_name"],
                data=np.c_[monthly_data["hr"].to_numpy()],
                attrs=monthly_data["hr"].attrs,
            ),
            "codec": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data["gtsm_tide_surge_slr"]],
                attrs=hourly_data["gtsm_tide_surge_slr"].attrs,
            ),
            "codec_tide": xr.DataArray(
                dims=["time", "station_name"],
                data=np.c_[hourly_data["gtsm_tide"]],
                attrs=hourly_data["gtsm_tide"].attrs,
            ),
            "distance": xr.DataArray(
                dims="station_name",
                data=[distance],
                attrs={
                    "long_name": "Distance between GESLA2 tide gauge station and closest HCC grid point",
                    "unit": "m",
                },
            ),
        },
        coords={
            "gesla_lat": xr.DataArray(dims="station_name", data=[tg_meta.latitude]),
            "gesla_lon": xr.DataArray(dims="station_name", data=[tg_meta.longitude]),
            "lat": xr.DataArray(dims="station_name", data=[hourly_data.lat.item()]),
            "lon": xr.DataArray(dims="station_name", data=[hourly_data.lon.item()]),
            "gesla_station_name": xr.DataArray(dims="station_name", data=[tg_f_name]),
            "station_name": xr.DataArray(dims="station_name", data=[tg_f_name]),
            "time": hourly_data.time,
            "annual_time": xr.DataArray(
                dims="annual_time",
                data=pd.date_range("1900", "2015", freq="AS"),
            ),
            "monthly_time": monthly_data.time.rename(time="monthly_time"),
        },
    )
    if len(amsl.annual_time) > 0:
        station_data["tg_amsl"] = amsl
    else:
        station_data["tg_amsl"] = xr.DataArray(
            np.full(
                (
                    len(station_data["annual_time"]),
                    len(station_data["station_name"]),
                ),
                np.nan,
            ),
            coords=[
                station_data["annual_time"],
                station_data["station_name"],
            ],
            dims=["annual_time", "station_name"],
        )
    station_data["tg_amsl_hourly"] = (
        station_data["tg_amsl"]
        .reindex(annual_time=station_data["time"].values)
        .rename(annual_time="time")
        .ffill(dim="time")
    )
    station_data["w_hour"] = gesla_hourly
    station_data["tg_w_hour_detr"] = (
        station_data["w_hour"] - station_data["tg_amsl_hourly"]
    )
    # val_ds_gesla_dict[station_name] = station_data
    station_data["hcc_surge"] = station_data["hcc"] - station_data["codec_tide"]
    station_data["hcc_cfact_surge"] = (
        station_data["hcc_cfact"] - station_data["codec_tide"]
    )
    return station_data


def main(conf):
    client = Client()
    print("started a dask.distributed client. Daskboard is on")
    print(client.dashboard_link)
    gesla_dir = conf.input_dir / "gesla3"
    tmp_output_dir = conf.output_dir / "tmp_gesla3_val_files"
    tmp_output_dir.mkdir(exist_ok=True)
    gtsm_index_all = pd.read_hdf(conf.preprocessed_input / "gtsm_index.h5")
    gtsm_index = gtsm_index_all[gtsm_index_all.coastal]
    gtsm_index_coords = np.array([gtsm_index.lat, gtsm_index.lon]).T
    bt_gtsm = BallTree(np.deg2rad(gtsm_index_coords), metric="haversine")

    meta_file = gesla_dir / "GESLA3_ALL 2.csv"
    tg_meta = pd.read_csv(meta_file)
    tg = tg_meta[
        (tg_meta["OVERALL RECORD QUALITY"] == "No obvious issues")
        & (tg_meta["GAUGE TYPE"] == "Coastal")
    ].sort_values(by="FILE NAME")
    gesla3 = GeslaDataset(
        meta_file=str(meta_file), data_path=str(gesla_dir) + "/GESLA3.0_ALL/"
    )

    datasets = Parallel(n_jobs=16)(
        delayed(save_station_netcdf)(
            tg_f_name, gesla3, bt_gtsm, gtsm_index, conf.output_dir, tmp_output_dir
        )
        for tg_f_name in tg["FILE NAME"]
    )
    datasets = [ds for ds in datasets if ds is not None]

    print("RAM Used (GB):", psutil.virtual_memory()[3] / 1000000000)
    print("concatenate station_data")
    val_ds_gesla = xr.concat(datasets, dim="station_name")

    del datasets
    print("RAM Used (GB):", psutil.virtual_memory()[3] / 1000000000)

    print("saving merged val_ds_gesla")
    val_ds_gesla.to_netcdf(conf.output_dir / "val_ds_extremes.nc")
    print(f"stored val_ds_extremes.nc at {conf.output_dir / 'val_ds_extremes.nc'}")


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
