import gc

import numpy as np
import pandas as pd
import psutil
import xarray as xr
from gslcomp.config import ConfigParser
from scipy.io import loadmat
from sklearn.neighbors import BallTree


def main(conf):
    if conf.gesla_2 == "public":
        raise NotImplementedError("Only implemented for gesla_2_unique")
    elif conf.gesla_2 == "unique":
        gesla_2_dir = conf.input_dir / "GESLA_2_unique"

    psmsl_monthly = xr.load_dataset(conf.input_dir / "psmsl_monthly")
    gtsm_index_all = pd.read_hdf(conf.preprocessed_input / "gtsm_index.h5")
    gtsm_index = gtsm_index_all[gtsm_index_all.coastal]
    gtsm_index_coords = np.array([gtsm_index.lat, gtsm_index.lon]).T
    bt_gtsm = BallTree(np.deg2rad(gtsm_index_coords), metric="haversine")

    val_ds_psmsl_dict = {}
    for i, (_, psmsl_i) in enumerate(psmsl_monthly.groupby("station_id")):
        search_point = np.array([[psmsl_i["lat"].item()], [psmsl_i["lon"].item()]]).T
        distance, index = bt_gtsm.query(np.deg2rad(search_point))
        earth_radius = 6371000
        distance = distance.item() * earth_radius
        gtsm_f_name = gtsm_index.iloc[index.item()].filename

        hcc = xr.load_dataset(
            conf.output_dir / "hourly_time_series" / f"{gtsm_f_name}.nc"
        )

        monthly_data = xr.load_dataset(
            conf.output_dir / "monthly_time_series" / f"{gtsm_f_name}.nc"
        )

        station_name = hcc.station_name.item()
        if station_name in val_ds_psmsl_dict:
            if (
                psmsl_i.msl.count().item()
                <= val_ds_psmsl_dict[station_name]["psmsl"].count().item()
            ):
                continue
        hcc_rel_long = xr.concat(
            [
                (
                    monthly_data["hr_relative_no_season"]
                    + monthly_data["seasonal_cycle"]
                ).loc[{"time": slice("1900", "1978")}],
                hcc["hcc_relative"].resample(time="MS", loffset="15d").mean(),
            ],
            dim="time",
        )
        hcc_abs_long = xr.concat(
            [
                (
                    monthly_data["hr_absolute_no_season"]
                    + monthly_data["seasonal_cycle"]
                ).loc[{"time": slice("1900", "1978")}],
                hcc["hcc_absolute"].resample(time="MS", loffset="15d").mean(),
            ],
            dim="time",
        )
        hcc_rel_cfact_long = xr.concat(
            [
                (
                    monthly_data["hr_relative_cfact_no_season"]
                    + monthly_data["seasonal_cycle"]
                ).loc[{"time": slice("1900", "1978")}],
                hcc["hcc_relative_cfact"].resample(time="MS", loffset="15d").mean(),
            ],
            dim="time",
        )
        hcc_abs_cfact_long = xr.concat(
            [
                (
                    monthly_data["hr_absolute_cfact_no_season"]
                    + monthly_data["seasonal_cycle"]
                ).loc[{"time": slice("1900", "1978")}],
                hcc["hcc_absolute_cfact"].resample(time="MS", loffset="15d").mean(),
            ],
            dim="time",
        )
        val_ds_psmsl_dict[station_name] = xr.Dataset(
            data_vars={
                "hcc": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc.hcc_relative.to_numpy()],
                    attrs={"long_name": "Hourly coastal water levels", "unit": "mm",},
                ),
                "hcc_cfact": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc.hcc_relative_cfact.to_numpy()],
                    attrs={
                        "long_name": "Counterfactual hourly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "geocentric_hcc": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc.hcc_absolute.to_numpy()],
                    attrs={
                        "long_name": "Geocentric hourly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "geocentric_hcc_cfact": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc.hcc_absolute_cfact.to_numpy()],
                    attrs={
                        "long_name": "Counterfactual geocentric hourly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "hcc_long": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[hcc_rel_long.to_numpy()],
                    attrs={"long_name": "Monthly coastal water levels", "unit": "mm",},
                ),
                "hcc_cfact_long": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[hcc_rel_cfact_long.to_numpy()],
                    attrs={
                        "long_name": "Counterfactual monthly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "geocentric_hcc_long": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[hcc_abs_long.to_numpy()],
                    attrs={
                        "long_name": "Geocentric monthly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "geocentric_hcc_cfact_long": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[hcc_abs_cfact_long.to_numpy()],
                    attrs={
                        "long_name": "Counterfactual geocentric monthly coastal water levels",
                        "unit": "mm",
                    },
                ),
                "hr": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[monthly_data["hr"].to_numpy()],
                    attrs=monthly_data["hr"].attrs,
                ),
                "codec": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc["gtsm_tide_surge_slr"]],
                    attrs=hcc["gtsm_tide_surge_slr"].attrs,
                ),
                "codec_tide": xr.DataArray(
                    dims=["time", "station_name"],
                    data=np.c_[hcc["gtsm_tide"]],
                    attrs=hcc["gtsm_tide"].attrs,
                ),
                "distance_psmsl": xr.DataArray(
                    dims="station_name",
                    data=[distance],
                    attrs={
                        "long_name": "Distance between PSMSL tide gauge station and closest HCC grid point",
                        "unit": "m",
                    },
                ),
                "psmsl": xr.DataArray(
                    dims=["monthly_time", "station_name"],
                    data=np.c_[psmsl_i.msl.to_numpy()],
                    coords={
                        "monthly_time": psmsl_i.rename(time="monthly_time").monthly_time
                    },
                    attrs={
                        "long_name": "Monthly water levels from the PSMSL tide gauge record",
                        "unit": "mm",
                    },
                ),
            },
            coords={
                "psmsl_lat": xr.DataArray(
                    dims="station_name", data=[psmsl_i.lat.item()]
                ),
                "psmsl_lon": xr.DataArray(
                    dims="station_name", data=[psmsl_i.lon.item()]
                ),
                "lat": xr.DataArray(dims="station_name", data=[hcc.lat.item()]),
                "lon": xr.DataArray(dims="station_name", data=[hcc.lon.item()]),
                "psmsl_station_name": xr.DataArray(
                    dims="station_name", data=[psmsl_i.station_name.item()]
                ),
                "psmsl_station_id": ("station_name", [psmsl_i.station_id.item()],),
                "station_name": xr.DataArray(
                    dims="station_name", data=[hcc.station_name.item()]
                ),
                "time": hcc.time,
                "annual_time": xr.DataArray(
                    dims="annual_time", data=pd.date_range("1900", "2015", freq="AS"),
                ),
                "monthly_time": monthly_data.time.rename(time="monthly_time"),
            },
        )

    val_ds_psmsl = xr.concat(val_ds_psmsl_dict.values(), dim="station_name")
    print("RAM Used (GB):", psutil.virtual_memory()[3] / 1000000000)
    del val_ds_psmsl_dict
    gc.collect()
    print("RAM Used (GB):", psutil.virtual_memory()[3] / 1000000000)

    val_ds_psmsl.to_netcdf(conf.output_dir / "val_ds_psmsl.nc")
    print(f"stored val_ds_psmsl.nc at {conf.output_dir / 'val_ds_psmsl.nc'}")


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
