from pathlib import Path

import xarray as xr
from gslcomp.config import ConfigParser


def general_postprocessing(raw_data):
    pp_data = raw_data.rename_dims(station_name="stations")
    pp_data = pp_data.drop_vars(["station_name", "is_dry"])
    # adjust long_name
    pp_data.time.attrs["long_name"] = "Time"
    pp_data.lat.attrs["long_name"] = "Latitude"
    pp_data.lon.attrs["long_name"] = "Longitude"
    # adjust axis
    pp_data.time.attrs["axis"] = "T"
    pp_data.lat.attrs["axis"] = "Y"
    pp_data.lon.attrs["axis"] = "X"
    # global attributes
    pp_data.attrs["references"] = (
        "The dataset is described in "
        "Treu et al. (2023) <https://doi.org/10.5194/essd-2023-112>. "
        "The source datasets are described in "
        "Dangendorf et al. (2019) <https://doi.org/10.1038/s41558-019-0531-8>, "
        "Muis et al. (2020) <https://doi.org/10.3389/fmars.2020.00263> "
        "and Oelsmann et al. (2023) <https://doi.org/10.5281/zenodo.8308347>"
    )
    pp_data.attrs["institution"] = "Potsdam Institute for Climate Impact Research (PIK)"
    pp_data.attrs[
        "project"
    ] = "Inter-Sectoral Impact Model Intercomparison Project phase 3a (ISIMIP3a)"
    pp_data.attrs[
        "contact"
    ] = "ISIMIP cross-sectoral science team <info@isimip.org> <https://www.isimip.org>"
    pp_data.attrs["grid_info"] = (
        "The irregular coastal grid points are distributed equidistantly "
        "along a smoothed global coastlines with "
        "a distance of 50 km globally and 10km in Europe "
        "with additional points added for tide gauge stations "
        "(described in Muis et all. 2019, https://doi.org/10.3389/fmars.2020.00263)."
    )
    return pp_data


def postprocess_factual_data(raw_data):
    pp_data = general_postprocessing(raw_data)
    pp_data = pp_data.rename(hcc="waterlevel")
    pp_data = pp_data.rename(hcc_geocentric="geocentricwaterlevel")
    # adjust long_name
    pp_data.waterlevel.attrs["long_name"] = "Relative Water Level"
    pp_data.geocentricwaterlevel.attrs["long_name"] = "Geocentric Water Level"
    # adjust comment
    water_level_comment = (
        "The relative coastal water level is derived as a combination of "
        "reconstructed geocentric water level based on Dangendorf et al. (2019), "
        "vertical land motion based on Oelsmann et al. (2023) and "
        "hourly storm-tide variations based on Muis et al. (2020). "
        "The data is aligned with the geocentric water level such "
        "that the last time step in the record is identical. It therefore describes "
        "the water level relative to a virtual tide gauge that is at the height of the "
        "geocentric water level at the last date in the record."
    )
    geocentric_water_level_comment = (
        "The geocentric coastal water level is "
        "independent of local vertical land motion. It is a combination of "
        "monthly water level reconstructions from Dangendorf et al. (2019) and "
        "hourly storm-tide variations based on Muis et al. (2020). The water "
        "level is standardized to a common vertical datum - the WGS 84 "
        "geoid, representing the absolute dynamic topography."
    )
    pp_data.waterlevel.attrs["comment"] = water_level_comment
    pp_data.geocentricwaterlevel.attrs["comment"] = geocentric_water_level_comment
    # adjust coordinates
    pp_data.waterlevel.encoding["coordinates"] = "lat lon"
    pp_data.geocentricwaterlevel.encoding["coordinates"] = "lat lon"
    return pp_data


def postprocess_counterfactual_data(raw_data):
    pp_data = general_postprocessing(raw_data)
    pp_data = pp_data.rename(hcc_cfact="waterlevel")
    pp_data = pp_data.rename(hcc_geocentric_cfact="geocentricwaterlevel")
    # adjust long_name
    pp_data.waterlevel.attrs["long_name"] = "Relative Water Level"
    pp_data.geocentricwaterlevel.attrs["long_name"] = "Geocentric Water Level"
    # adjust comment
    water_level_comment = (
        "The counterfactual relative coastal water level "
        "is derived by removing the quadratic trend since 1900 from the "
        "factual relative coastal water level. The factual data is a "
        "combination of reconstructed geocentric water level data based on "
        "Dangendorf et al. (2019), vertical land motion (VLM) based on Oelsmann "
        "et al. (2023) and hourly storm-tide variations based on Muis et al. (2020)."
    )

    geocentric_water_level_comment = (
        "The counterfactual geocentric coastal water level "
        "is derived by removing the quadratic trend since 1900 from "
        "the factual geocentric coastal water level. The factual geocentric "
        "coastal water level is a combination of "
        "monthly water level reconstructions from Dangendorf et al. (2019) and "
        "hourly storm-tide variations based on Muis et al. (2020)."
    )
    pp_data.waterlevel.attrs["comment"] = water_level_comment
    pp_data.geocentricwaterlevel.attrs["comment"] = geocentric_water_level_comment

    # adjust coordinates
    pp_data.waterlevel.encoding["coordinates"] = "lat lon"
    pp_data.geocentricwaterlevel.encoding["coordinates"] = "lat lon"
    return pp_data


def main(conf):
    hourly_station_data_dir = conf.output_dir / "hourly_time_series_annual_split"
    obsclim_merged_output = (
        conf.output_dir / "merged_hourly_time_series" / "obsclim" / "global" / "hourly"
    )

    obsclim_merged_output.mkdir(exist_ok=True, parents=True)
    counterclim_merged_output = (
        conf.output_dir
        / "merged_hourly_time_series"
        / "counterclim"
        / "global"
        / "hourly"
    )
    counterclim_merged_output.mkdir(exist_ok=True, parents=True)
    for year in range(1979, 2016):
        # merge factual data
        factual_data = []
        for file in sorted(
            (hourly_station_data_dir / f"{year}").glob(f"factual_{year}_*.nc")
        ):
            factual_data.append(xr.load_dataset(file))
        merged_factual_data = xr.concat(factual_data, dim="station_name")
        pp_factual_data = postprocess_factual_data(merged_factual_data)
        pp_factual_data = pp_factual_data.transpose("time", "stations")
        # store relative
        pp_factual_relative = pp_factual_data.drop_vars(["geocentricwaterlevel"])
        pp_factual_relative.attrs["title"] = "Hourly coastal relative water levels"

        pp_factual_relative.to_netcdf(
            obsclim_merged_output / f"hcc_obsclim_waterlevel_global_hourly_{year}.nc"
        )
        # store geocentric
        pp_factual_geocentric = pp_factual_data.drop_vars(["waterlevel"])
        pp_factual_geocentric.attrs["title"] = "Hourly coastal geocentric water levels"
        pp_factual_geocentric.to_netcdf(
            obsclim_merged_output
            / f"hcc_obsclim_geocentricwaterlevel_global_hourly_{year}.nc"
        )

        # merge counterfactual data
        counterfactual_data = []
        for file in sorted(
            (hourly_station_data_dir / f"{year}").glob(f"counterfactual_{year}_*.nc")
        ):
            counterfactual_data.append(xr.load_dataset(file))
        merged_counterfactual_data = xr.concat(counterfactual_data, dim="station_name")
        pp_cfact_data = postprocess_counterfactual_data(merged_counterfactual_data)
        pp_cfact_data = pp_cfact_data.transpose("time", "stations")
        # store relative
        pp_cfact_relative = pp_cfact_data.drop_vars(["geocentricwaterlevel"])
        pp_cfact_relative.attrs[
            "title"
        ] = "Counterfactual hourly coastal relative water levels"
        pp_cfact_relative.to_netcdf(
            counterclim_merged_output
            / f"hcc_counterclim_waterlevel_global_hourly_{year}.nc"
        )
        # store geocentric
        pp_cfact_geocentric = pp_cfact_data.drop_vars(["waterlevel"])
        pp_cfact_geocentric.attrs[
            "title"
        ] = "Counterfactual hourly coastal geocentric water levels"
        pp_cfact_geocentric.to_netcdf(
            counterclim_merged_output
            / f"hcc_counterclim_geocentricwaterlevel_global_hourly_{year}.nc"
        )


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
