import numpy as np
import xarray as xr
from gslcomp.config import ConfigParser


def general_postprocessing(raw_data):
    pp_data = raw_data.sel(time=slice("1900", "2015"))
    pp_data = pp_data.drop_vars(["station_name"])
    pp_data = pp_data.drop_dims(["2"])
    pp_data = pp_data.drop_dims(
        [dim for dim in pp_data.dims if dim.startswith("string")]
    )

    pp_data.time.attrs["long_name"] = "Time"
    pp_data.lat.attrs["long_name"] = "Latitude"
    pp_data.lon.attrs["long_name"] = "Longitude"
    # adjust axis
    pp_data.time.attrs["axis"] = "T"
    pp_data.lat.attrs["axis"] = "Y"
    pp_data.lon.attrs["axis"] = "X"
    # global attributes
    pp_data.attrs["references"] = (
        "The dataset is described in "
        "Treu et al. (2023) <https://doi.org/10.5194/essd-2023-112>. "
        "The source datasets are described in "
        "Dangendorf et al. (2019) <https://doi.org/10.1038/s41558-019-0531-8>, "
        "Muis et al. (2020) <https://doi.org/10.3389/fmars.2020.00263> "
        "and Oelsmann et al. (2023) <https://doi.org/10.5281/zenodo.8308347>"
    )
    pp_data.attrs["institution"] = "Potsdam Institute for Climate Impact Research (PIK)"
    pp_data.attrs[
        "project"
    ] = "Inter-Sectoral Impact Model Intercomparison Project phase 3a (ISIMIP3a)"
    pp_data.attrs[
        "contact"
    ] = "ISIMIP cross-sectoral science team <info@isimip.org> <https://www.isimip.org>"
    pp_data.attrs["grid_info"] = (
        "The irregular coastal grid points are distributed equidistantly "
        "along a smoothed global coastlines with "
        "a distance of 50 km globally and 10km in Europe "
        "with additional points added for tide gauge stations "
        "(described in Muis et all. 2019, https://doi.org/10.3389/fmars.2020.00263)."
    )
    return pp_data


def postprocess_factual_data(raw_data):
    pp_data = general_postprocessing(raw_data)
    pp_data = pp_data[
        [
            "hr_relative_no_season",
            "hr_absolute_no_season",
            "seasonal_cycle",
            "lat",
            "lon",
        ]
    ]
    pp_data["waterlevel"] = pp_data["hr_relative_no_season"] + pp_data["seasonal_cycle"]
    pp_data.waterlevel.attrs["long_name"] = "Relative Water Level"
    pp_data["geocentricwaterlevel"] = (
        pp_data["hr_absolute_no_season"] + pp_data["seasonal_cycle"]
    )
    pp_data.geocentricwaterlevel.attrs["long_name"] = "Geocentric Water Level"
    water_level_comment = (
        "This variable represents the monthly relative coastal water level as a "
        "combination of reconstructed geocentric water level data based on "
        "Dangendorf et al. (2019) and vertical land motion (VLM) based on Oelsmann "
        "et al. (2023). "
        "The data is aligned with the HCC hourly relative coastal water levels."
    )
    geocentric_water_level_comment = (
        "This variable represents the monthly geocentric coastal water level that is "
        "independent of local vertical land motion. It is based on "
        "monthly water level reconstructions from Dangendorf et al. (2019) "
        "excluding the effects of vertical land motion."
    )
    pp_data.waterlevel.attrs["comment"] = water_level_comment
    pp_data.geocentricwaterlevel.attrs["comment"] = geocentric_water_level_comment
    pp_data = pp_data.drop_vars(
        ["hr_absolute_no_season", "hr_relative_no_season", "seasonal_cycle"]
    )
    # adjust coordinates
    pp_data.waterlevel.encoding["coordinates"] = "lat lon"
    pp_data.geocentricwaterlevel.encoding["coordinates"] = "lat lon"
    # assert for all variables that there is no NaN data
    # drop the _FillValue attribute
    for data_var in pp_data.data_vars:
        assert len(pp_data[data_var].dropna(dim="time")) == len(pp_data[data_var])
        try:
            del pp_data[data_var].encoding["_FillValue"]
        except KeyError:
            continue
    for var in ["lat", "lon"]:
        assert not np.isnan(pp_data[var].values).any()
        try:
            del pp_data[var].encoding["_FillValue"]
        except KeyError:
            continue
    return pp_data


def postprocess_counterfactual_data(raw_data):
    pp_data = general_postprocessing(raw_data)
    pp_data = pp_data[
        [
            "hr_relative_cfact_no_season",
            "hr_absolute_cfact_no_season",
            "seasonal_cycle",
            "lat",
            "lon",
        ]
    ]
    # create waterlevel and geocentricwaterlevel variables
    pp_data["waterlevel"] = (
        pp_data["hr_relative_cfact_no_season"] + pp_data["seasonal_cycle"]
    )
    pp_data["geocentricwaterlevel"] = (
        pp_data["hr_absolute_cfact_no_season"] + pp_data["seasonal_cycle"]
    )
    # adjust attributes
    pp_data.waterlevel.attrs["long_name"] = "Relative Water Level"
    pp_data.geocentricwaterlevel.attrs["long_name"] = "Geocentric Water Level"
    water_level_comment = (
        "This variable represents the counterfactual "
        "monthly coastal relative water level. "
        "It is derived by removing the quadratic trend since 1900 from the "
        "factual monthly coastal relative water level. The factual data is a "
        "combination of reconstructed geocentric water level data based on "
        "Dangendorf et al. (2019) and "
        "vertical land motion (VLM) based on Oelsmann et al. (2023)."
    )

    geocentric_water_level_comment = (
        "This variable represents the counterfactual "
        "monthly geocentric coastal water level. "
        "It is derived by removing the quadratic trend since 1900 from "
        "the factual geocentric coastal water level. The factual geocentric "
        "coastal water level is based on "
        "monthly water level reconstructions from Dangendorf et al. (2019) "
        "excluding the effects from vertical land motion."
    )
    pp_data.waterlevel.attrs["comment"] = water_level_comment
    pp_data.geocentricwaterlevel.attrs["comment"] = geocentric_water_level_comment
    pp_data = pp_data.drop_vars(
        ["hr_absolute_cfact_no_season", "hr_relative_cfact_no_season", "seasonal_cycle"]
    )
    # adjust coordinates
    pp_data.waterlevel.encoding["coordinates"] = "lat lon"
    pp_data.geocentricwaterlevel.encoding["coordinates"] = "lat lon"
    for data_var in pp_data.data_vars:
        assert len(pp_data[data_var].dropna(dim="time")) == len(pp_data[data_var])
        try:
            del pp_data[data_var].encoding["_FillValue"]
        except KeyError:
            continue
    for var in ["lat", "lon"]:
        assert not np.isnan(pp_data[var].values).any()
        try:
            del pp_data[var].encoding["_FillValue"]
        except KeyError:
            continue
    return pp_data


def main(conf):
    monthly_data_dir = conf.output_dir / "monthly_time_series"
    obsclim_merged_monthly_dir = conf.output_dir / "merged_monthly_time_series_zenodo"

    obsclim_merged_monthly_dir.mkdir(exist_ok=True, parents=True)
    counterclim_merged_monthly_dir = (
        conf.output_dir / "merged_monthly_time_series_zenodo"
    )

    counterclim_merged_monthly_dir.mkdir(exist_ok=True, parents=True)
    factual_data = []
    counterfactual_data = []
    for file in sorted(monthly_data_dir.iterdir()):
        raw_data = xr.load_dataset(file)
        pp_factual_data = postprocess_factual_data(raw_data)
        pp_counterfactual_data = postprocess_counterfactual_data(raw_data)
        factual_data.append(pp_factual_data)
        counterfactual_data.append(pp_counterfactual_data)

    # merge the single DataArrays
    merged_factual_data = xr.concat(factual_data, dim="stations")
    merged_factual_data = merged_factual_data.transpose("time", "stations")
    merged_counterfactual_data = xr.concat(counterfactual_data, dim="stations")
    merged_counterfactual_data = merged_counterfactual_data.transpose(
        "time", "stations"
    )
    # store individual netcdf files for each variable
    merged_factual_relative_wl = merged_factual_data.drop_vars(["geocentricwaterlevel"])
    merged_factual_relative_wl.attrs["title"] = "Monthly coastal relative water levels"
    merged_factual_relative_wl.to_netcdf(
        obsclim_merged_monthly_dir
        / "hcc_obsclim_waterlevel_global_monthly_1900_2015.nc"
    )
    merged_factual_geocentric_wl = merged_factual_data.drop_vars(["waterlevel"])
    merged_factual_geocentric_wl.attrs[
        "title"
    ] = "Monthly coastal geocentric water levels"
    merged_factual_geocentric_wl.to_netcdf(
        obsclim_merged_monthly_dir
        / "hcc_obsclim_geocentricwaterlevel_global_monthly_1900_2015.nc",
    )
    merged_counterfactual_relative_wl = merged_counterfactual_data.drop_vars(
        ["geocentricwaterlevel"]
    )
    merged_factual_relative_wl.attrs[
        "title"
    ] = "Counterfactual monthly coastal relative water levels"
    merged_counterfactual_relative_wl.to_netcdf(
        counterclim_merged_monthly_dir
        / "hcc_counterclim_waterlevel_global_monthly_1900_2015.nc"
    )
    merged_counterfactual_geocentric_wl = merged_counterfactual_data.drop_vars(
        ["waterlevel"]
    )
    merged_counterfactual_geocentric_wl.attrs[
        "title"
    ] = "Counterfactual monthly coastal geocentric water levels"
    merged_counterfactual_geocentric_wl.to_netcdf(
        counterclim_merged_monthly_dir
        / "hcc_counterclim_geocentricwaterlevel_global_monthly_1900_2015.nc"
    )
    print(f"stored monthly files in {monthly_data_dir}")


if __name__ == "__main__":
    main(conf=ConfigParser().parse())
