"""hr is provided as a .mat file (Matlab file format).
This file relies on the order of GTSM stations.
Index i in the .mat is the i`th gtsm station when they are ordered lexicographically.
To make sure nothing gets mixed up, we compare lat and lon attributes of gtsm and hr stations and
save the hr data in single netCDF files for each station with the gtsm file name.
"""
import datetime
from pathlib import Path

import numpy as np
from pytest import approx
from scipy.io import loadmat

import xarray as xr
from cftime import date2num
from gslcomp.computation import load_gtsm_output
from gslcomp.config import ConfigParser, logging_setup, save_args
from gslcomp.data_loading import get_input_filenames
from gslcomp.utils import get_available_slurm_tasks, run_parallelized


def center_month(time_float):
    """
    Convert monthly time index represented as floats to datetime objects. The index is at the center of the month.
    Parameters
    ----------
    time_float: iterable(float)
        contains floats where the integer part represents the year and
        the float represents the month with Jan=0, Feb=1/12, Mar=2/12 ..., Dec=11/12

    Returns
    -------
    datetime_list: list(datetime)
        a list of datetime objects for the 16th day of the month specified in the float input

    """
    return [
        datetime.datetime(year=int(t), month=round(t % 1 * 12 + 1), day=16)
        for t in time_float
    ]


def beginning_month(time_float):
    """
    Convert monthly time index represented as floats to datetime objects. The index is at the beginning of the month.

    Parameters
    ----------
    time_float: iterable(float)
        contains floats where the integer part represents the year and
        the float represents the month with Jan=0, Feb=1/12, Mar=2/12 ..., Dec=11/12

    Returns
    -------
    datetime_list: list(datetime)
        a list of datetime objects for the 1st day of the month specified in the float input

    """
    return [
        datetime.datetime(year=int(t), month=round(t % 1 * 12 + 1), day=1)
        for t in time_float
    ]


def hr_mat_to_netcdf(index, filenames, msl_aviso, mdt_global, conf):
    """
    Extract monthly hr data for single station from .mat file and save in netCDF format.
    The time series indices are transformed to pandas datetime indices. Currently, this implementation
    is based on the assumption that
    hr and gtsm have the same ordering of files. This assumption is checked
    by asserting equal lat and lon
    Parameters
    ----------
    index : int
        station to be saved
    filenames : list()
        filenames of gtsm locations
    msl_aviso : xr.DataArray
        aviso satellite altimetry
    mdt_global : xr.DataArray
        mean dynamic topography
    conf : argparse.Namespace
        experiment specific configurations
    Returns
    -------

    """
    gtsm_tide_slr = load_gtsm_output(
        conf.input_dir / "gtsm_tide_surge_slr" / f"{filenames[index]}.nc", conf
    )
    hr_mat = loadmat(conf.input_dir / Path("HybridRec_Upd0422.mat"))

    standard_attrs = {
        "references": "https://doi.org/10.1038/s41558-019-0531-8",
        "units": "mm",
        "positive": "up",
    }

    hr_xr = xr.Dataset(
        data_vars={
            "hr": xr.DataArray(
                dims="time",
                data=hr_mat["HRC"][index, :],
                attrs={
                    "long_name": "Hybrid reconstruction water levels",
                    "comment": "Coastal water levels from the hybrid reconstrution (HR) dataset. It contains"
                    "vertical land motion from GIA and from the crustal deformation from present day ice melt",
                    **standard_attrs,
                },
            ),
            "hr_no_gia": xr.DataArray(
                dims="time",
                data=hr_mat["HRCnoGIA"][index, :],
                attrs={
                    "long_name": "Hybrid reconstruction water levels without GIA",
                    "comment": "hr without vertical land motion from GIA",
                    **standard_attrs,
                },
            ),
            "hr_no_season": xr.DataArray(
                dims="time",
                data=hr_mat["HRCnoseason"][index, :],
                attrs={
                    "long_name": "Hybrid reconstruction water levels without seasonal cycle",
                    "comment": "HR without seasonal cycle",
                    **standard_attrs,
                },
            ),
            "hr_no_gia_no_season": xr.DataArray(
                dims="time",
                data=hr_mat["HRCnoseasonnoGIA"][index, :],
                attrs={
                    "long_name": "Hybrid reconstrution water levels without seasonal cycle and GIA",
                    "comment": "Similar to hr but without the seasonal cycle and without vertical land motion from GIA",
                    **standard_attrs,
                },
            ),
            "seasonal_cycle": xr.DataArray(
                dims="time",
                data=hr_mat["SCC"][index, :],
                attrs={
                    "long_name": "Seasonal cycle",
                    "comment": "Average monthly seasonal cycle from satellite altimetry",
                    **standard_attrs,
                },
            ),
            "Crustal_PD": xr.DataArray(
                dims="time",
                data=np.append(
                    np.full(
                        shape=hr_mat["tTG"].shape[0] - hr_mat["tALT"].shape[0],
                        fill_value=np.nan,
                    ),
                    hr_mat["Crustal_PD"][index, :],
                ),
                attrs={
                    "long_name": "Crustal deformation from present day ice melt",
                    "comment": "Vertical land motion due to the crustal deformation from present day ice melt. "
                    "Positive means uplift of land, negative is subsidence of land. " 
                    "The influence on relative sea level has the opposite sign",
                    "references": "https://doi.org/10.1038/s41558-019-0531-8",
                    "units": "mm",
                    "positive": "down",
                },
            ),
            "hr_altimetry_with_gia_and_crustal_pd_no_season": xr.DataArray(
                dims="time",
                data=np.append(
                    np.full(
                        shape=hr_mat["tTG"].shape[0] - hr_mat["tALT"].shape[0],
                        fill_value=np.nan,
                    ),
                    hr_mat["ALTnoseason"][index, :],
                ),
                attrs={
                    "long_name": "Satellite altimetry with GIA and CrustalPD",
                    "comment": "Data from AVISO used in https://doi.org/10.1038/s41558-019-0531-8 for comparison. "
                    "Crustal_PD and GIA is added to the AVISO altimetry such that it can be compared to HR."
                    "The seasonal cycle is removed.",
                    **standard_attrs,
                },
            ),
        },
        coords={
            "time": xr.DataArray(
                dims="time",
                data=date2num(
                    center_month(hr_mat["tTG"][:, 0]), "hours since 1900-01-01 00:00:00"
                ),
                attrs={
                    "standard_name": "time",
                    "long_name": "time of measurement",
                    "units": "hours since 1900-01-01 00:00:00",
                    "missing_value": -999.9,
                    "bounds": "time_bounds",
                },
            ),
            "time_bounds": xr.DataArray(
                dims=["time", "2"],
                data=date2num(
                    np.array(
                        [
                            beginning_month(hr_mat["tTG"][:, 0]),
                            beginning_month(hr_mat["tTG"][:, 0] + 1 / 12),
                        ],
                    ).transpose(),
                    "hours since 1900-01-01 00:00:00",
                ),
                attrs={
                    "standard_name": "time_bounds",
                    "long_name": "bounds of one time interval",
                    "units": "hours since 1900-01-01 00:00:00",
                    "missing_value": -999.9,
                },
            ),
            "lat": xr.DataArray(
                data=hr_mat["LC"][index, 1],
                attrs={
                    "standard_name": "latitude",
                    "long_name": "latitude of the observation",
                    "units": "degrees_north",
                },
            ),
            "lon": xr.DataArray(
                data=hr_mat["LC"][index, 0],
                attrs={
                    "standard_name": "longitude",
                    "long_name": "longitude of the observation",
                    "units": "degrees_east",
                },
            ),
            "station_name": xr.DataArray(
                data=gtsm_tide_slr.station_name,
                attrs={
                    "standard_name": "station_name",
                    "long_name": "station name",
                    "f_role": "timeseries_id",
                },
            ),
        },
        attrs={
            "Convention": "CF-1.9",
            "featureType": "timeSeries",
            "title": "Hybrid reconstruction of coastal waterlevels",
            "reference": "https://doi.org/10.1038/s41558-019-0531-8",
        },
    )

    # Ensure hardcoded tALT and tTG overlap
    np.testing.assert_almost_equal(
        hr_mat["tALT"][:, 0], hr_mat["tTG"][-hr_mat["tALT"].shape[0] :, 0]
    )
    assert hr_xr.lat == approx(gtsm_tide_slr["latitude"], abs=0.001)
    assert hr_xr.lon == approx(gtsm_tide_slr["longitude"], abs=0.001)
    
    hr_xr["msl_aviso"] = data=msl_aviso.sel(latitude=hr_xr["lat"], longitude=hr_xr["lon"], method="nearest").item()
    hr_xr["mdt_global"] = data=mdt_global.sel(latitude=hr_xr["lat"], longitude=hr_xr["lon"], method="nearest").item()

    save_dir = conf.input_dir / "hr_monthly"
    save_dir.mkdir(parents=True, exist_ok=True)
    hr_xr.to_netcdf(save_dir / f"{filenames[index]}.nc", format="NETCDF4_CLASSIC")


def main(conf):
    """
    Split hybrid response .mat file to single hdf files for each station.
    Computation is parallelized on cluster.

    Parameters
    ----------
    conf : argparse.Namespace
        experiment specific configurations
    Returns
    -------

    """
    print(f"Start transforming hybrid response .mat file to single stations")

    n_tasks, task_id = get_available_slurm_tasks()
    logger = logging_setup(conf.output_dir)
    filenames = get_input_filenames(conf.input_dir)
    aviso = xr.load_dataset(conf.input_dir / "adt_y1993-2019_fill_dis.nc")
    aviso["longitude"] = (aviso.longitude + 180) % 360 - 180
    aviso = aviso.sortby("longitude")
    msl_aviso = aviso.adt.loc[{"time": slice("1993", "2015")}].mean(dim="time") * 1000
    mdt_global = xr.load_dataset(
        conf.input_dir / "mdt_cnes_cls18_global_nn_interpolated.nc"
    )["mdt"]

    run_parallelized(
        worker_id=task_id,
        n_workers=n_tasks,
        n_jobs=len(filenames),
        logger=logger,
        func=hr_mat_to_netcdf,
        kwargs={
            "filenames": filenames,
            "msl_aviso": msl_aviso,
            "mdt_global": mdt_global,
            "conf": conf,
        },
    )


if __name__ == "__main__":
    parsed_conf = ConfigParser().parse()
    save_args(vars(parsed_conf), parsed_conf.output_dir / "args.txt")
    main(conf=parsed_conf)
