"""
Preprocessing for vlm data from W. Hammond. The data from hammond is provided for the CoDEC locations.

Input:  vlm data from a .txt file with a simple tabular format.
Output: netCDF file following cf conventions for point data.
        http://cfconventions.org/Data/cf-conventions/cf-conventions-1.9/cf-conventions.html#point-data
"""

import pandas as pd
import netCDF4 as nc
import numpy as np
from gslcomp.config import ConfigParser, save_args
from pathlib import Path


def main(conf):
    vlm_df = pd.read_csv(
        conf.input_dir / Path("GPS_VLM_at_Coasts_LongTable.csv"),
        delimiter=";",
        names=[
            "longitude",
            "latitude",
            "vertical rate estimate from GPS",
            "su",
            "eta",
            "zeta",
            "upsilon",
        ],
        # engine="python",
        dtype=np.float64,
    )

    with nc.Dataset(
        conf.input_dir / "GPS_VLM.nc", mode="w", format="NETCDF4_CLASSIC"
    ) as ncfile:
        # dimensions
        station_dim = ncfile.createDimension("station", len(vlm_df))
        # auxiliary dimension variables
        lat = ncfile.createVariable("lat", np.float32, ("station",))
        lat.standard_name = "latitude"
        lat.long_name = "latitude of the observation"
        lat.units = "degrees_north"
        lat[:] = vlm_df["latitude"]

        lon = ncfile.createVariable("lon", np.float32, ("station",))
        lon.standard_name = "longitude"
        lon.long_name = "longitude of the observation"
        lon.units = "degrees_east"
        lon[:] = vlm_df["longitude"]

        # data variables
        vlm_rate = ncfile.createVariable("vlm_rate", np.float32, ("station",))
        vlm_rate.standard_name = "vlm_rate"
        vlm_rate.long_name = "vertical rate estimate from GPS"
        vlm_rate.coordinates = "lat lon"
        vlm_rate.units = "mm / y"
        vlm_rate.positive = "up"
        vlm_rate[:] = vlm_df["vertical rate estimate from GPS"]

        su = ncfile.createVariable("su", np.float32, ("station",))
        su.standard_name = "su"
        su.long_name = "formal uncertainty of rate as if were a mean of rates from nearby stations (even though it is a median)"
        su.coordinates = "lat lon"
        su[:] = vlm_df["su"]

        eta = ncfile.createVariable("eta", np.float32, ("station",))
        eta.standard_name = "eta"
        eta.long_name = (
            "s the median value of the spatial structure function given the distance"
        )
        eta.coordinates = "lat lon"
        eta.units = "1"
        eta[:] = vlm_df["eta"]

        zeta = ncfile.createVariable("zeta", np.float32, ("station",))
        zeta.standard_name = "zeta"
        zeta.long_name = "spatial variability in rate from GPS network"
        zeta.coordinates = "lat lon"
        zeta[:] = vlm_df["zeta"]

        upsilon = ncfile.createVariable("upsilon", np.float32, ("station",))
        upsilon.standard_name = "upsilon"
        upsilon.long_name = "temporal variability of rate from GPS time series"
        upsilon.coordinates = "lat lon"
        upsilon[:] = vlm_df["upsilon"]

        # global attributes
        ncfile.setncatts(
            {
                "Convention": "CF-1.8",
                "featureType": "point",
                "title": "Linear rate of vertical land motion estimated from GPS at coastal grid points",
                "source": "GPS",
                "reference": "Hammond, W.C., G. Blewitt, C. Kreemer, S. Nerem, 2021, Global vertical land motion for studies of sea level rise, J. Geophys. Res. - Solid Earth, 126(7), e2021JB022355, https://doi.org/10.1029/2021JB022355",
            }
        )


if __name__ == "__main__":
    parsed_conf = ConfigParser().parse()
    save_args(vars(parsed_conf), parsed_conf.output_dir / "args.txt")
    main(conf=parsed_conf)
