""" The netCDF file from hammond et al. that can be downloaded at http://geodesy.unr.edu/vlm.php needs to be
reformatted such that it can be properly used by xarray and cdo. Then CDO is used to interpolate NaN values, such that
we have data for our grid points that are close to the coast
"""
from pathlib import Path
import xarray as xr
from gslcomp.config import ConfigParser, save_args
import subprocess


def main(conf):
    """
    re-formats vlm netcdf file and interpolates it with nearest neighbour interpolation
    Parameters
    ----------
    conf: argparse.Namespace
        experiment specific configurations. Only input_dir and name must be given.

    Returns
    -------

    """
    vlm_hammond = xr.load_dataset(conf.input_dir / Path("VLM_Global_Imaged.nc"))

    data = vlm_hammond["VU"][:-1, :].transpose().data
    vu = xr.DataArray(
        data,
        coords={
            "lat": vlm_hammond["LatGrid"][0, :].data,
            "lon": vlm_hammond["LonGrid"][:-1, 0].data,
        },
        attrs={"standard_name": "vu", "long_name": "vu", "units": "mm/yr"},
    )
    vu["lat"].attrs = {
        "standard_name": "lat",
        "long_name": "Latitude",
        "units": "degrees_north",
        "axis": "Y",
    }
    vu["lon"].attrs = {
        "standard_name": "lon",
        "long_name": "Longitude",
        "units": "degrees_east",
        "axis": "X",
    }

    vlm = xr.Dataset(
        {"VU": vu},
        attrs={
            "geospatial_lat_max": vu["lat"][-1].item(),
            "geospatial_lat_min": vu["lat"][0].item(),
            "geospatial_lat_resolution": 0.25,
            "geospatial_lat_units": "degrees_north",
            "geospatial_lon_max": vu["lon"][-1].item(),
            "geospatial_lon_min": vu["lon"][0].item(),
            "geospatial_lon_resolution": 0.25,
            "geospatial_lon_units": "degrees_east",
        },
    )
    vlm.to_netcdf(
        path=conf.input_dir / "vlm_hammond.nc",
        encoding={
            "lon": {"dtype": "float64", "_FillValue": "NaN"},
            "lat": {"dtype": "float64", "_FillValue": "NaN"},
        },
    )

    # Interpolate with CDO
    cmd = f"cdo setmisstonn {conf.input_dir}/vlm_hammond.nc {conf.input_dir}/vlm_hammond_interpolated.nc"
    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError:
        cmd = "module load cdo && " + cmd
        subprocess.check_call(cmd, shell=True)


if __name__ == "__main__":
    parsed_conf = ConfigParser().parse()
    save_args(vars(parsed_conf), parsed_conf.output_dir / "args.txt")
    main(conf=parsed_conf)
