from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from gslcomp.config import ConfigParser, save_args
from sklearn.neighbors import BallTree


def main(conf):
    oelsmann_vlm = xr.load_dataset(
        conf.input_dir
        / "VLM_reconstruction_draft_data_Oelsmann_2022"
        / "VLM_reconstruction_v2.nc"
    )

    # get index
    coor_coastal = xr.load_dataset(conf.input_dir / "coor_coastal.nc")
    oelsmann_vlm_coords = np.array(
        [oelsmann_vlm["lat"].as_numpy(), oelsmann_vlm["lon"].as_numpy()]
    ).transpose()
    coor_coastal_coords = np.array(
        [
            coor_coastal.station_y_coordinate.to_numpy(),
            coor_coastal.station_x_coordinate.to_numpy(),
        ]
    ).transpose()
    #    kd_oelsmann_vlm = KDTree(oelsmann_vlm_coords)
    bt_oelsmann_vlm = BallTree(np.deg2rad(oelsmann_vlm_coords), metric="haversine")

    # _, index = kd_oelsmann_vlm.query(coor_coastal_coords)
    _, index = bt_oelsmann_vlm.query(np.deg2rad(coor_coastal_coords))

    full_oelsmann_vlm = oelsmann_vlm["Full_VLM_reconstruction_mean"][:, index.flatten()]
    # todo this is only a bad bug fix. should do this properly
    full_oelsmann_vlm[0][full_oelsmann_vlm[0].isnull()] = full_oelsmann_vlm[1][
        full_oelsmann_vlm[0].isnull()
    ]
    full_oelsmann_vlm.name = "vlm"

    time = xr.DataArray(
        pd.date_range(start="1900", end="1995-05", freq="MS") + pd.DateOffset(days=15),
        dims="time",
    )
    x_monthly = (time.dt.year + (time.dt.month - 1) / 12).to_numpy()

    oelsmann_vlm_lin = xr.DataArray(
        dims=["station_name", "time"],
        data=oelsmann_vlm.VLM_trend_coefficient_mean[index.flatten()].to_numpy()[
            :, None
        ]
        * x_monthly[None, :]
        * 1000,
        coords={"time": time, "station_name": coor_coastal.station_name.to_numpy()},
        name="vlm",
    )
    lin_trend = xr.DataArray(
        dims=["station_name", "time"],
        data=(
            oelsmann_vlm_lin.to_numpy()
            - oelsmann_vlm_lin[:, -1].to_numpy()[:, None]
            + full_oelsmann_vlm[0, :].to_numpy()[:, None] * 1000
        ),
        coords={
            "station_name": coor_coastal.station_name.to_numpy(),
            "time": oelsmann_vlm_lin.time,
            "lat": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lat.to_numpy()
            ),
            "lon": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lon.to_numpy()
            ),
        },
        name="vlm",
    )
    annual_vlm_part = xr.DataArray(
        dims=["time", "station_name"],
        data=full_oelsmann_vlm.to_numpy() * 1000,
        coords={
            "station_name": coor_coastal.station_name.to_numpy(),
            "time": pd.DatetimeIndex(full_oelsmann_vlm.time)
            - pd.offsets.DateOffset(days=198),
            "lat": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lat.to_numpy()
            ),
            "lon": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lon.to_numpy()
            ),
        },
        name="vlm",
    )

    monthly_vlm_part = annual_vlm_part.resample(time="MS").asfreq()
    monthly_vlm_part["time"] = pd.DatetimeIndex(
        monthly_vlm_part.time
    ) + pd.offsets.DateOffset(days=15)
    monthly_vlm_part = monthly_vlm_part.interpolate_na(dim="time", method="linear")
    oelsmann_vlm_monthly = xr.combine_by_coords(
        [
            lin_trend,
            monthly_vlm_part,
        ]
    )["vlm"]

    oelsmann_vlm_annual = oelsmann_vlm_monthly.resample(time="1a").mean()

    print("storing vlm files ...")
    oelsmann_vlm_annual.to_netcdf(
        conf.input_dir / "oelsmann_vlm_annual.nc", format="NETCDF4_CLASSIC"
    )
    oelsmann_vlm_monthly.to_netcdf(
        conf.input_dir / "oelsmann_vlm_monthly.nc", format="NETCDF4_CLASSIC"
    )
    print("vlm_files_stored")
    return oelsmann_vlm_annual, oelsmann_vlm_monthly


if __name__ == "__main__":
    parsed_conf = ConfigParser().parse()
    save_args(vars(parsed_conf), parsed_conf.output_dir / "args.txt")
    main(conf=parsed_conf)
