"""Helper functions that are not specific to one part of the module"""
import os
import re
import traceback
from math import ceil

import numpy as np
import pandas as pd
import xarray as xr
from func_timeout import FunctionTimedOut, func_timeout
from scipy.io import loadmat
from scipy.spatial import KDTree
from sklearn.neighbors import BallTree


def h5store(filename, df, **kwargs):
    """
    Store pandas DataFrame + metadata to hdf5 file.
    Taken from
    https://stackoverflow.com/questions/29129095/save-additional-attributes-in-pandas-dataframe/29130146#29130146

    Parameters
    ----------
    filename : str, pathlib.Path
        file path to hdf5 file
    df : pandas.DataFrame
        DataFrame to save
    kwargs : dict
        metadata to save

    Returns
    -------
    """
    store = pd.HDFStore(filename)
    store.put("mydata", df)
    if not kwargs is None:
        store.get_storer("mydata").attrs.metadata = kwargs
    store.close()


def h5load(filename):
    """
    Loads pandas DataFrame + metadata from hdf5 file.
    Taken from
    https://stackoverflow.com/questions/29129095/save-additional-attributes-in-pandas-dataframe/29130146#29130146

    Parameters
    ----------
    filename : str, pathlib.Path
        file path to hdf5 file.

    Returns
    -------
    data, metadata : pd.DataFrame, dict
        data is a DataFrame, metadata is a dictionary containing metadata
    """
    if not filename.is_file():
        raise FileNotFoundError(f"there is no such file: {filename}")
    with pd.HDFStore(filename, "r") as store:
        data = store["mydata"]
        try:
            metadata = store.get_storer("mydata").attrs.metadata
        except KeyError:
            metadata = None
    return data, metadata


def get_available_slurm_tasks():
    """
    if on cluster: get number of available tasks on the cluster
    for parallel computations and the current task_id.
    Otherwise n_tasks is 1 and the task_id is 0 for no parallelization.
    Returns
    -------
    n_tasks : int
        available tasks for parallelization
    task_id : int
        current task id
    """
    try:
        task_id = int(os.environ["SLURM_ARRAY_TASK_ID"])
        n_tasks = int(os.environ["SLURM_ARRAY_TASK_COUNT"])
    except KeyError:
        task_id = 0
        n_tasks = 1
    return n_tasks, task_id


def run_parallelized(worker_id, n_workers, n_jobs, logger, func, kwargs, timeout=2000):
    """
    Distributes the number of total_jobs to the available workers for parallelization.
    For each job func is run with kwargs as arguments and an additional index argument.
    If an error is raised, it is written to error.log and the other jobs are continued.
    Parameters
    ----------
    worker_id : int
        id of the current worker (SLURM_ARRAY_TASK_ID)
    n_workers : int
        total number of workers (SLURM_ARRAY_TASK_COUNT)
    n_jobs : int
        total number of jobs
    logger : logging.logger
        error log
    func : Any
        function to be executed in parallel
    kwargs : dict
        inputs to the function
    timeout : int, optional
        func is interrupted after the specified timeout in seconds.
        Defaults to 2000 seconds.

    Returns
    -------
    errors: [Error]
        list of errors that were raised by the parallel jobs
    """
    jobs_per_array = ceil(n_jobs / n_workers)
    start = worker_id * jobs_per_array
    end = min(n_jobs, (worker_id + 1) * jobs_per_array)
    print(
        f"this is task {worker_id} running {end - start} jobs from job number {start} to job number {end - 1}"
    )
    errors = []
    for i in range(start, end):
        print(f"This is job number {i}")
        kwargs["index"] = i
        # noinspection PyBroadException
        try:
            func_timeout(timeout, func, kwargs=kwargs)
        except FunctionTimedOut as error:
            print(f"Calculating job {i} timed out after {timeout} seconds.")
            logger.error(f"Job {i} error : {error}")
            logger.error(traceback.format_exc())
            errors.append(error)
            continue
        except Exception as error:
            print(traceback.format_exc())
            logger.error(traceback.format_exc())
            errors.append(error)
            continue
    return errors


def get_lat_lon(filename, gesla_2):
    """
    Parse lat and lon from GESLA2 files
    Parameters
    ----------
    filename: pathlib.Path
        the path of the GESLA2 file
    gesla_2: {"public", "unique"}
        version of gesla
    Returns
    -------
    lat, lon: float, float
        latitude and longitude of the station
    """
    if gesla_2 == "public":
        with open(filename, errors="ignore") as f:
            for line in f:
                if "LATITUDE" in line:
                    lat = float(re.findall(r"[+-]?\d+\.\d+", line)[0])
                    break
            for line in f:
                if "LONGITUDE" in line:
                    lon = float(re.findall(r"[+-]?\d+\.\d+", line)[0])
                    break
        return lat, lon
    elif gesla_2 == "unique":
        f = loadmat(str(filename))
        lat = f["Lat"].item()
        lon = f["Lon"].item()
        return lat, lon
    else:
        raise NotImplementedError(
            f"configuration gesla_2={gesla_2} is not implemented. choose one of ['public', 'unique']"
        )


def read_tide_gauge(file, gesla_2):
    """
    read tide_gauge data from a GESLA2 file
    Parameters
    ----------
    file: pathlib.Path
        the path of the GESLA2 file
    gesla_2: {"public", "unique"}
        version of gesla

    Returns
    -------
    sl: pd.Series
        Time series with the hourly sea-level
    """
    if gesla_2 == "public":
        sl = pd.read_csv(
            file,
            delim_whitespace=True,
            header=0,
            skiprows=32,
            parse_dates=[[0, 1]],
            names=["date", "time", "sl", "qc_flag", "used-in-extreme-analysis_flag"],
        )
        sl = sl.set_index("date_time")
        # although there is a null value encoded in the files, it seemed to always be -99.999
        # but for some locations, e.g. zanzibar, null_values are encoded as -99.99 hence we do not read null_value
        # from the file but use it hardcoded instead
        sl.loc[sl["sl"] < -99, "sl"] = np.nan
        #
        sl.loc[sl["qc_flag"] >= 3, "sl"] = np.nan
        return sl["sl"]

    elif gesla_2 == "unique":
        tide_gauge = loadmat(str(file))
        # The value 719529 is the (matlab) datenum value of the Unix epoch start (1970-01-01),
        # which is the default origin for pd.to_datetime()
        t = pd.to_datetime(tide_gauge["Thour"][:, 0] - 719529, unit="D").round("H")
        whour_detr = pd.Series(tide_gauge["Whour_detr"][:, 0], index=t)
        amsl = pd.Series(
            tide_gauge["AMSL"][:, 1],
            index=pd.to_datetime(tide_gauge["AMSL"][:, 0].astype(int).astype(str)),
        )

        tg_df = pd.DataFrame({"whour_detr": whour_detr, "amsl": amsl}, index=t)
        sl = tg_df["whour_detr"] + tg_df["amsl"].interpolate(method="ffill").fillna(0)
        return sl
    else:
        raise NotImplementedError(
            f"configuration gesla_2={gesla_2} is not implemented. choose one of ['public', 'unique']"
        )


def get_k_nearest_neighbours(locations_gdf, lat, lon, k):
    """
    returns a GeoDataFrame with the k nearest neighbours in locations_gdf for the given lat,lon
    it is sorted by increasing distance

    Parameters
    ----------
    locations_gdf: geopandas.GeoDataFrame
        must have Point geometry
    lat: float
    lon: float
    k: int

    Returns
    -------
    neighbours: geopandas.GeoDataFrame
        k nearest neighbours sorted by increasing distance
    """
    if k > len(locations_gdf):
        k = len(locations_gdf)
    locations = list(locations_gdf.geometry.apply(lambda x: (x.x, x.y)))
    btree = KDTree(locations)
    dist, idx = btree.query((lon, lat), k=k)
    if k == 1:
        idx = [idx]
    neighbours = locations_gdf.iloc[idx].reset_index(drop=True)
    neighbours["dist"] = dist
    return neighbours


def remove_annual_mean(timeseries):
    """

    Parameters
    ----------
    timeseries: pd.Series
        a series with a date_time index. Column name of the index should not be nan


    Returns
    -------
    pd.Series
    """
    timeseries.index.name = "date_time"
    df = pd.DataFrame({"sl": timeseries, "year": timeseries.index.year})
    df = (
        df.reset_index()
        .merge(df.resample("1a").mean(), on="year", how="left")
        .set_index(timeseries.index.name)
    )
    return df["sl_x"] - df["sl_y"]


def rescale_and_normalize(timeseries, time_scale, vertical_alignment, daily_max=False):
    """
    rescales the timeseries with mean values over the specified bin size (time_scale). Additionally the timeseries is
    normalized with the method specified in vertical_alignment
    Parameter
    ----------
    timeseries: pd.Series
    time_scale: str
    vertical_alignment: {'no', 'annual', 'reference_period'}
        method for normalization

    Returns
    -------
    pd.Series
    """

    # vertical alignment
    if vertical_alignment == "no":
        aligned_ts = timeseries
    elif vertical_alignment == "annual":
        aligned_ts = remove_annual_mean(timeseries)
    elif vertical_alignment == "reference_period":
        ref_period_mean = timeseries["1979-01-01":"1984-12-31"].mean()
        # note that ref_period_mean might be nan if there is no data. Then we return nan for the rescaled timeseries
        if np.isnan(ref_period_mean):
            raise ValueError("no data in the reference period")
        else:
            aligned_ts = timeseries - ref_period_mean
    else:
        raise NotImplementedError(
            f"vertical alignment method {vertical_alignment} is not implemented"
        )

    rescaled_ts = (
        aligned_ts.resample("1d").max()
        if time_scale == "1dmax"
        else aligned_ts.resample(time_scale).mean()
    )

    return rescaled_ts


def select_stations(val_nc, min_obs_years):
    n_obs_years_after_1993 = (
        val_nc["tg_amsl"]
        .loc[{"annual_time": slice("1993", "2015")}]
        .count(dim="annual_time")
    )
    sel_station_names = val_nc["station_name"].where(
        n_obs_years_after_1993 >= min_obs_years, drop=True
    )
    return val_nc.sel(station_name=sel_station_names)


def select_stations_and_align(val_nc, min_obs_years):
    val_nc = select_stations(val_nc, min_obs_years)

    alignment_offset = (
        val_nc["hcc_rel"]
        .loc[{"time": slice("1993", "2015")}]
        .resample(time="AS")
        .mean()
        .rename(time="annual_time")
        - val_nc["tg_amsl"].loc[{"annual_time": slice("1993", "2015")}]
    ).mean(dim="annual_time")
    val_nc["tg_amsl_aligned"] = val_nc["tg_amsl"] + alignment_offset

    val_nc["tg_hourly_waterlevel_aligned"] = (
        val_nc["tg_hourly_waterlevel"] + alignment_offset
    )
    return val_nc


def lons_360_to_neg180_180(lon):
    lons = []
    for l in lon:
        lons.append(l) if l <= 180 else lons.append(l - 360)
    return np.array(lons)


def get_reg_mask(file):
    reg_mask_to_sorted_basins = {
        1: 3,
        2: 6,
        3: 4,
        4: 1,
        5: 2,
        6: 5,
        7: 0,
    }

    reg_mask = loadmat(file)
    # Split Subpolar North Atlantic into Eastern and Western parts
    rm = np.where(
        (reg_mask["reg"] == 4)
        & (lons_360_to_neg180_180(reg_mask["Lon"].flatten()) > -40).reshape(72, 36),
        7,
        reg_mask["reg"],
    )

    reg_mask_sorted = np.array(
        [
            [
                reg_mask_to_sorted_basins[rm.T[i, j]]
                if not np.isnan(rm.T[i, j])
                else np.nan
                for j in range(rm.T.shape[1])
            ]
            for i in range(rm.T.shape[0])
        ]
    )

    reg_mask_xr = xr.Dataset(
        {
            "reg_mask": xr.DataArray(
                data=reg_mask_sorted,
                coords={
                    "lat": reg_mask["Lat"][0, :],
                    "lon": lons_360_to_neg180_180(reg_mask["Lon"][:, 0]),
                },
                dims=["lat", "lon"],
            ).sortby("lon")
        }
    )

    # change shape from a lat lon grid to station_data and drop points without a defined basin
    lats, lons = np.meshgrid(reg_mask_xr.lat, reg_mask_xr.lon)
    reg_mask_2 = xr.DataArray(
        data=reg_mask_xr.reg_mask.to_numpy().T.flatten(),
        dims="idx",
        coords={
            "idx": np.arange(len(reg_mask_xr.reg_mask.to_numpy().flatten())),
            "lat": xr.DataArray(dims="idx", data=lats.flatten()),
            "lon": xr.DataArray(dims="idx", data=lons.flatten()),
        },
    )

    return reg_mask_xr, reg_mask_2.dropna(dim="idx")


def get_coastline(extremes, reg_mask):
    index = get_index_of_nearest_locations(ds=reg_mask, target_ds=extremes)
    return xr.DataArray(
        data=reg_mask[index].to_numpy(),
        dims="station_name",
        coords={"station_name": extremes.station_name},
    )


def get_index_of_nearest_locations(ds, target_ds):
    kd = KDTree(np.array([ds["lat"].as_numpy(), ds["lon"].as_numpy()]).transpose())
    _, index = kd.query(
        np.array(
            [target_ds["lat"].as_numpy(), target_ds["lon"].as_numpy()]
        ).transpose(),
        k=1,
    )
    return index
