"""Generates a composition of the GTSM and HR datasets following the given configuration
options"""
import xarray as xr

from gslcomp.computation import combine_gtsm_hr
from gslcomp.config import ConfigParser, logging_setup, save_args
from gslcomp.data_loading import get_input_filenames
from gslcomp.utils import (get_available_slurm_tasks, lons_360_to_neg180_180,
                           run_parallelized)


def main(conf):
    """
    Generates a composition of the GTSM and HR datasets following the given configuration options.

    Parameters
    ----------
    conf : argparse.Namespace
        experiment specific configurations

    Returns
    -------
    errors: [Error]
        list of errors that were raised by the parallel jobs
    """
    print(f"Starting {conf.name}")
    n_tasks, task_id = get_available_slurm_tasks()
    logger = logging_setup(conf.output_dir, conf.overwrite)
    filenames = get_input_filenames(conf.input_dir)
    vlm_oelsmann_ds = xr.load_dataset(conf.input_dir / "oelsmann_vlm_monthly.nc")
    vlm_hammond = xr.load_dataset(conf.input_dir / "GPS_VLM.nc")
    frederikse_2020_total = xr.load_dataset(conf.input_dir / "total.nc")
    frederikse_2020_tws = xr.load_dataset(conf.input_dir / "tws.nc")
    sed_frederikse_2020 = (
        frederikse_2020_total.total_rad_mean - frederikse_2020_tws.tws_rad_mean
    )
    sed_frederikse_2020["lon"] = lons_360_to_neg180_180(sed_frederikse_2020.lon)

    errors = run_parallelized(
        worker_id=task_id,
        n_workers=n_tasks,
        n_jobs=len(filenames),
        logger=logger,
        func=combine_gtsm_hr,
        timeout=conf.timeout,
        kwargs={
            "filenames": filenames,
            "conf": conf,
            "vlm_oelsmann_ds": vlm_oelsmann_ds,
            "vlm_hammond": vlm_hammond,
            "sed_frederikse_2020": sed_frederikse_2020.sortby("lon"),
        },
    )
    return errors


if __name__ == "__main__":
    parsed_conf = ConfigParser().parse()
    save_args(vars(parsed_conf), parsed_conf.output_dir / "args.txt")
    main(conf=parsed_conf)
