import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt


class BarPlot():
    def __init__(self, data, norm, cmap, sort_basins):
        self.x = np.arange(len(data.station_name))
        self.lat = data["lat"]
        self.coastline = data["coastline"]
        self.counts = np.cumsum(
        [(ds["coastline"].count()) for _, ds in data.groupby("coastline")]
        )
        self.norm = norm
        self.cmap = cmap
        self.sort_basins = sort_basins
    
    
    def plot(self, panel, fname):
        plot_sizes = [1] * len(panel) + [0.5]
        f = plt.figure(figsize=(9.75, 2.5 * sum(plot_sizes)))
        gridscale = int(np.prod([1 / s for s in set(plot_sizes) if s < 1]))
        axs = [
            plt.subplot2grid(
                (int(sum(plot_sizes) * gridscale), 1),
                (i * gridscale, 0),
                rowspan=int(s * gridscale),
            )
            for i, s in enumerate(plot_sizes)
        ]
                
        for i, panel in enumerate(panel):
            panel["fn"](self.x, axs[i], **panel["kwargs"])

        for ax in axs[:-1]:
            ax.set_xticks([], [])
            ax.set_xlim([0, self.x[-1]])
            self._set_axis_layout(ax, y_grid=True)
        self._set_axis_layout(axs[-1], y_grid=False)
        
        for i, ax in enumerate(axs):
            ax.text(0.05, 0.85, f'({chr(ord("a") + i)})', transform=ax.transAxes, weight="bold")
    

            
        ax = axs[-1]
        ylims = -79.99, 79.99  # remove uppermost label 80 to decrease paddin
        ax.set_ylim(ylims)

        ax.set_ylabel("Latitude (°N)")
        y = self.lat

        m = mpl.cm.ScalarMappable(norm=self.norm, cmap=self.cmap)
        ax.bar(self.x, y, color=m.to_rgba(self.coastline), alpha=0.3, width=1)

        # Basin labels
        dx = -(self.counts - np.concatenate([np.array([0]), self.counts[:-1]]))/2
        y = ylims[0] + 0.01 * (ylims[1] - ylims[0])

        basin_label = {
            "Subpolar North Atl. East": " Subpolar \n North \n Atl. East ",
            "Subpolar North Atl. West": " Subpolar \n North \n Atl. West ",
            "Subtropical North Atlantic": "Subtropical\nNorth\nAtlantic",
            # "Northwest Pacific": "Northwest\nPacific",
            "South Atlantic": "South\nAtlantic",
            "Indian Ocean - South Pacific": "Indian Ocean\nSouth Pacific"
        }

        label_offset = {"(5) Subtropical N. Atlantic": 30, "(3) East Pacific": 30}

        for i, basin in enumerate(self.sort_basins):
            ax.text(
                self.counts[i] + dx[i],
                y + label_offset.get(basin, 0),
                basin_label.get(basin, basin),
                fontdict={"ha": "center", "fontsize": "x-small", "va": "bottom"},
            )


        f.set_dpi(400)
        plt.tight_layout(h_pad=0)
        plt.savefig(fname, dpi=300)
    
    
    def _set_axis_layout(self, ax, y_grid):
        for c in self.counts[:-1]:
            ax.axvline(c - 0.5, color="gray", linewidth=0.5, alpha=0.7)

        ax.set_xlim(self.x[0] - 0.5, xmax=self.x[-1] + 0.5)

        # y-grid
        if y_grid:
            ax.yaxis.grid(alpha=0.7)
            
def plot_rmse(x, ax, y1, y2, y1_label, y2_label, c1, c2, ylabel, ylim=None):
    kwargs = {"marker": ".", "s": 10, "alpha": 0.8, "linewidths": 1}
    ax.scatter(x, y1, label=y1_label, color=c1, **kwargs)
    ax.scatter(x, y2, label=y2_label, color=c2, **kwargs)

    ax.bar(
        x[y1 <= y2],
        y2[y1 <= y2] - y1[y1 <= y2],
        bottom=y1[y1 <= y2],
        color=c1,
        width=1,
        alpha=0.6,
    )

    ax.bar(
        x[y1 > y2],
        y1[y1 > y2] - y2[y1 > y2],
        bottom=y2[y1 > y2],
        color=c2,
        width=1,
        alpha=0.6,
    )
    ax.set_ylabel(ylabel)

    if not (ylim is None):
        ax.set_ylim(ylim)

    ax.legend(loc="upper right")