from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from scipy.io import loadmat
from scipy.spatial import KDTree
from sklearn.neighbors import BallTree


def load_region_info():
    sheets = [
        "Subpolar North Atlantic",
        "Indian Ocean - South Pacific",
        "Subtropical North Atlantic",
        "East Pacific",
        "South Atlantic",
        "Northwest Pacific",
    ]

    dfs = []
    for i, basin in enumerate(sheets):
        sheet = pd.read_excel(
            Path("../input") / "region_info.xlsx", sheet_name=i
        ).set_index("Unnamed: 0")
        sheet.index = sheet.index + 1000 * i
        sheet["Basin"] = basin
        if basin == "Subpolar North Atlantic":
            lon = sheet["Longitude"].values
            coastline = np.where(
                (lon < 50) | (lon > 320),
                "Subpolar North Atl. East",
                "Subpolar North Atl. West",
            )
        else:
            coastline = basin
        sheet["coastline"] = coastline
        dfs.append(sheet)
    region_info = pd.concat(dfs)
    duplicate_indices = [119, 120]
    region_info = region_info.loc[
        [ix for i, ix in enumerate(region_info.index) if ix not in duplicate_indices]
    ]
    return region_info


def get_oelsmann_vlm(oelsmann_vlm, global_monthly):
    index = get_index_of_nearest_locations(ds=oelsmann_vlm, target_ds=global_monthly)

    full_oelsmann_vlm = oelsmann_vlm["Full_VLM_reconstruction_mean"][:, index]
    # todo this is only a bad bug fix. should do this properly
    full_oelsmann_vlm[0][full_oelsmann_vlm[0].isnull()] = full_oelsmann_vlm[1][
        full_oelsmann_vlm[0].isnull()
    ]
    full_oelsmann_vlm.name = "vlm"

    time = global_monthly.time.loc[:"1995-05-16"]

    x_monthly = (time.dt.year + (time.dt.month - 1) / 12).to_numpy()

    oelsmann_vlm_lin = xr.DataArray(
        dims=["station_name", "time"],
        data=oelsmann_vlm.VLM_trend_coefficient_mean[index].to_numpy()[:, None]
        * x_monthly[None, :]
        * 1000,
        coords={"time": time, "station_name": global_monthly.station_name},
        name="vlm",
    )
    lin_trend = xr.DataArray(
        dims=["station_name", "time"],
        data=(
            oelsmann_vlm_lin.to_numpy()
            - oelsmann_vlm_lin[:, -1].to_numpy()[:, None]
            + full_oelsmann_vlm[0, :].to_numpy()[:, None] * 1000
        ),
        coords={
            "station_name": global_monthly.station_name,
            "time": oelsmann_vlm_lin.time,
            "lat": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lat.to_numpy()
            ),
            "lon": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lon.to_numpy()
            ),
        },
        name="vlm",
    )
    annual_vlm_part = xr.DataArray(
        dims=["time", "station_name"],
        data=full_oelsmann_vlm.to_numpy() * 1000,
        coords={
            "station_name": global_monthly.station_name,
            "time": pd.DatetimeIndex(full_oelsmann_vlm.time)
            - pd.offsets.DateOffset(days=198),
            "lat": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lat.to_numpy()
            ),
            "lon": xr.DataArray(
                dims="station_name", data=full_oelsmann_vlm.lon.to_numpy()
            ),
        },
        name="vlm",
    )
    oelsmann_vlm_long = xr.combine_by_coords(
        [
            lin_trend,
            annual_vlm_part,
        ]
    )["vlm"]

    monthly_vlm_part = annual_vlm_part.resample(time="MS").asfreq()
    monthly_vlm_part["time"] = pd.DatetimeIndex(
        monthly_vlm_part.time
    ) + pd.offsets.DateOffset(days=15)
    monthly_vlm_part = monthly_vlm_part.interpolate_na(dim="time", method="linear")
    oelsmann_vlm_monthly = xr.combine_by_coords(
        [
            lin_trend,
            monthly_vlm_part,
        ]
    )["vlm"]
    oelsmann_vlm_annual = oelsmann_vlm_monthly.resample(time="1a").mean()
    return oelsmann_vlm_annual, oelsmann_vlm_monthly


def get_neighbors(psmsl_annual, extremes):
    psmsl_annual_coords = np.array([psmsl_annual.lat, psmsl_annual.lon]).T
    bt = BallTree(np.deg2rad(psmsl_annual_coords), metric="haversine")
    search_coords = np.array([extremes.lat, extremes.lon]).T
    distance, index = bt.query(np.deg2rad(search_coords))
    earth_radius = 6371

    return xr.Dataset(
        data_vars={
            "distance": xr.DataArray(
                dims="gesla_station_name",
                data=distance.flatten() * earth_radius,
                attrs={
                    "long_name": "distance between this instance and its neighbour",
                    "units": "km",
                },
            )
        },
        coords={
            "gesla_station_name": xr.DataArray(
                dims="gesla_station_name", data=extremes.station_name.to_numpy()
            ),
            "gesla_lat": xr.DataArray(
                dims="gesla_station_name", data=extremes.lat.to_numpy()
            ),
            "gesla_lon": xr.DataArray(
                dims="gesla_station_name", data=extremes.lon.to_numpy()
            ),
            "psmsl_station_id": xr.DataArray(
                dims="gesla_station_name",
                data=psmsl_annual.station_id[index.flatten()].to_numpy(),
            ),
            "psmsl_station_name": xr.DataArray(
                dims="gesla_station_name",
                data=psmsl_annual.station_name[index.flatten()].to_numpy(),
            ),
            "psmsl_lat": xr.DataArray(
                dims="gesla_station_name",
                data=psmsl_annual.lat[index.flatten()].to_numpy(),
            ),
            "psmsl_lon": xr.DataArray(
                dims="gesla_station_name",
                data=psmsl_annual.lon[index.flatten()].to_numpy(),
            ),
        },
    )


def rmse(a, b, dim):
    return np.sqrt(((a - b) ** 2).mean(dim=dim))
